from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class FrozensetTests(TranspileTestCase):
    def test_creation(self):
        # Empty dict
        self.assertCodeExecution("""
            x = frozenset()
            print(x)

            x = frozenset({'a'})
            print(x)

            y = {1,2}
            x = frozenset(y)
            print(x)
            y = {3}
            print(x)

            y = ['c']
            x = frozenset(y)
            print(x)
            y.append('d')
            print(x)

            y = (1,2)
            x = frozenset(y)
            print(x)
            y = ()
            print(x)

            y ='a'
            x = frozenset(y)
            print(x)
            y = y + 'b'
            print(x)

            """)

    def test_contains(self):
        # Normal Contains
        self.assertCodeExecution("""
            x = frozenset()
            print(1 in x)

            x = frozenset({'a'})
            print('a' in x)
            print('b' in x)

            y = {1,2}
            x = frozenset(y)
            print(2 in x)
            print(3 in x)

            y = ['c']
            x = frozenset(y)
            print('c' in x)
            print('a' in x)

            y = (1,2)
            x = frozenset(y)
            print(3 in x)
            print(1 in x)

            y = 'a'
            x = frozenset(y)
            print('e' in x)
            print('a' in x)

            """)

    def test_not_contains(self):
        # Normal Not Contains
        self.assertCodeExecution("""
            x = frozenset()
            print(1 in x)

            x = frozenset({'a'})
            print('b' not in x)
            print('a' not in x)

            y = {1,2}
            x = frozenset(y)
            print(4 not in x)
            print(1 not in x)

            y = ['c']
            x = frozenset(y)
            print('d' not in x)
            print('c' not in x)

            y = (1,2)
            x = frozenset(y)
            print(1 not in x)
            print(3 not in x)

            y = 'a'
            x = frozenset(y)
            print('a' not in x)
            print('d' not in x)

            """)

    def test_iteration(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = frozenset()
            #We are not printing each element because python and voc store the items in different ways.
            for s in x:
                print(x.__contains__(s))
            for s in y:
                print(y.__conatins__(s))

            """)

    def test_copy(self):
        self.assertCodeExecution("""
            x = frozenset()
            y = x.copy()
            print(x is y)

            x = frozenset({'a'})
            y = x.copy()
            print(x is y)

            y = {1,2}
            x = frozenset(y)
            z = x.copy()
            print(x is z)

            y = ['c']
            x = frozenset(y)
            z = x.copy()
            print(x is z)

            y = (1,2)
            x = frozenset(y)
            z = x.copy()
            print(x is z)

            y = 'a'
            x = frozenset(y)
            z = x.copy()
            print(x is z)

            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = frozenset("hello world")
            try:
                x[0] = "goodbye"
            except TypeError as err:
                print(err)
        """)

    def test_isdisjoint(self):
        self.assertCodeExecution("""
            x = frozenset("hello world")
            y = set("hello")
            z = frozenset()
            t = "hello"
            w = 2.0

            print(x.isdisjoint(y))
            print(x.isdisjoint(z))

            # iterable test
            print(x.isdisjoint(t))

            # not-iterable test
            try:
                print(x.isdisjoint(w))
            except TypeError as err:
                print(err)
            """)

    def test_issubset(self):
        self.assertCodeExecution("""
            a = frozenset('abc')
            b = frozenset('abcde')
            c = set()
            d = 'wxyz'

            print(a.issubset(b))
            print(a.issubset(a))
            print(a.issubset(c))

            # iterable test
            print(a.issubset(d))
            """)

        # not-iterable test
        self.assertCodeExecution("""
            a = frozenset({1, 2, 3})
            b = 1
            try:
                print(a.issubset(b))
            except TypeError as err:
                print(err)
            """)

    def test_issuperset(self):
        self.assertCodeExecution("""
            a = frozenset('abcd')
            b = frozenset('ab')
            c = set()
            d = 'ab'

            print(a.issuperset(b))
            print(a.issuperset(a))
            print(a.issuperset(c))

            # iterable test
            print(a.issuperset(d))
            """)

        # not-iterable test
        self.assertCodeExecution("""
            a = frozenset({1, 2, 3})
            b = 1
            try:
                print(a.issuperset(b))
            except TypeError as err:
                print(err)
            """)

    def test_union(self):
        self.assertCodeExecution("""
            x = frozenset({1, 2, 3})
            y = frozenset({3, 4, 5})
            z = [5, 6, 7]
            w = 1
            t = frozenset()

            print(x.union(y))

            # empty set test
            print(x.union(t))

            # multiple args test
            print(x.union(y, z))

            # iterable test
            print(x.union(z))

            # not-iterable test
            try:
                print(x.union(w))
            except TypeError as err:
                print(err)
            """)

    def test_intersection(self):
        self.assertCodeExecution("""
            x = frozenset({1, 2, 3})
            y = frozenset({2, 3, 4})
            z = [3, 6, 7]
            w = 1
            t = frozenset()

            print(x.intersection(y))

            # empty set test
            print(x.intersection(t))

            # multiple args test
            print(x.intersection(y, z))

            # iterable test
            print(x.intersection(z))

            # not-iterable test
            try:
                print(x.intersection(w))
            except TypeError as err:
                print(err)
            """)

    def test_difference(self):
        self.assertCodeExecution("""
            x = frozenset({1, 2, 3})
            y = frozenset({2, 3, 4})
            z = [1, 6]
            w = 1
            t = frozenset()

            print(x.difference(y))

            # empty set test
            print(x.difference(t))

            # multiple args test
            print(x.difference(y, z))

            # iterable test
            print(x.difference(z))

            # not-iterable test
            try:
                print(x.difference(w))
            except TypeError as err:
                print(err)
            """)

    def test_symmetric_difference(self):
        self.assertCodeExecution("""
            x = frozenset({1, 2, 3})
            y = frozenset({2, 3, 4})
            z = [1, 6, 7]
            w = 1
            t = frozenset()

            print(x.symmetric_difference(y))

            # empty set test
            print(x.symmetric_difference(t))

            # iterable test
            print(x.symmetric_difference(z))

            # not-iterable test
            try:
                print(x.symmetric_difference(w))
            except TypeError as err:
                print(err)

            """)


class UnaryFrozensetOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'frozenset'


class BinaryFrozensetOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'frozenset'


class InplaceFrozensetOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'frozenset'



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class MinTests(TranspileTestCase):
    def test_args(self):
        self.assertCodeExecution("""
            try:
                min()
            except TypeError:
                print("Threw an error as expected")

            # Single integer
            try:
                min(4)
            except TypeError:
                print("Threw an error as expected")

            # Multiple integers
            print(min(1, 5, 2, 4))

            # Mixed types
            print(min(2, 5.0, True))
            print(min(5.0, 2, True))
            print(min(True, 5.0, 2))

            # String (an iterable)
            print(min("Hello World"))

            # Multiple strings
            print(min("Hello World", "Goodbye World"))
            """)

    def test_iterable(self):
        self.assertCodeExecution("""
            # Empty iterable
            try:
                min([])
            except ValueError:
                print("Threw an error as expected")

            # Single iterable argument
            print(min([1, 5, 2, 4]))

            # Multiple iterables
            print(min([1, 6], [1, 5, 2, 4]))
            """)

    def test_default(self):
        self.assertCodeExecution("""
            # Empty iterable
            print(min([], default=42))

            # Single iterable argument
            print(min([1, 5, 2, 4], default=42))

            # Multiple iterables
            try:
                print(min([1, 6], [1, 5, 2, 4], default=42))
            except TypeError:
                print("Threw an error as expected")
            """)

    def test_key(self):
        self.assertCodeExecution("""
            # key applied over args
            print(min(51, 42, 33, 24, key=lambda v: v % 10))

            # key applied over iterable
            print(min([51, 42, 33, 24], key=lambda v: v % 10))

            print(min([(2, 4), (5, 0), (4, 0), (9, 3)], key=lambda t: t[1]))
            """)


class BuiltinMinFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["min"]

    not_implemented = [
        'test_tuple',
    ]



from .. utils import TranspileTestCase  # , BuiltinFunctionTestCase


class InputTests(TranspileTestCase):
    pass


# FIXME: This test can't run without a redirection for stdin.
# class BuiltinInputFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
#     functions = ["input"]

#     not_implemented = [
#         'test_bool',
#         'test_bytearray',
#         'test_bytes',
#         'test_class',
#         'test_complex',
#         'test_dict',
#         'test_float',
#         'test_frozenset',
#         'test_int',
#         'test_list',
#         'test_None',
#         'test_NotImplemented',
#         'test_range',
#         'test_set',
#         'test_slice',
#         'test_str',
#         'test_tuple',
#     ]



import os
import sys

from ..java import (
    Class as JavaClass,
    Classref as JavaClassref,
    Code as JavaCode,
    Method as JavaMethod,
    SourceFile as JavaSourceFile,
    opcodes as JavaOpcodes
)
from .blocks import Block, Accumulator
from .methods import (
    CO_GENERATOR, GeneratorFunction, Function,
)
from .types import java, python
from .types.primitives import (
    ALOAD_name, ASTORE_name, free_name,
)
# from .debug import DEBUG


class Module(Block):
    def __init__(self, namespace, sourcefile, verbosity=0):
        super().__init__(verbosity=verbosity)
        self.sourcefile = sourcefile

        parts = os.path.splitext(sourcefile)[0].split(os.path.sep)
        if parts[-1] == '__init__':
            parts.pop()
            self.has_init_file = True
        else:
            self.has_init_file = False

        # If the sourcefile started with a /, the first part will
        # be an empty string; replace that with the namespace.
        # Otherwise, prepend the namespace to the parts.
        if parts[0] == '':
            parts[0] = namespace
        else:
            parts = [namespace] + parts

        self.namespace = '.'.join(parts[:-1])
        self.name = parts[-1]

        self.functions = []
        self.classes = []
        self.anonymous_inner_class_count = 0

        # Preallocate space for self when
        # the module static block is invoked
        self.parameters = [None]
        self.local_vars['self'] = 0

        self.store_module()

    @property
    def module(self):
        return self

    @property
    def full_name(self):
        return '.'.join(self.namespace.split('.')[1:] + [self.name])

    @property
    def descriptor(self):
        return '/'.join(self.namespace.split('.') + [self.name])

    @property
    def class_name(self):
        return '.'.join(self.namespace.split('.') + ([self.name, '__init__'] if self.has_init_file else [self.name]))

    @property
    def class_descriptor(self):
        return '/'.join(self.namespace.split('.') + ([self.name, '__init__'] if self.has_init_file else [self.name]))

    def build_child_class_descriptor(self, child_name):
        return '/'.join([self.descriptor, child_name])

    def store_module(self):
        # Stores the current module as a local variable
        if ('#module') not in self.local_vars:
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),

                ASTORE_name('#module'),
            )

    def store_name(self, name, declare=False):
        self.add_opcodes(
            ASTORE_name('#value'),

            ALOAD_name('#module'),

            ALOAD_name('#value'),
            python.Object.set_attr(name),

            free_name('#value')
        )

    def store_dynamic(self):
        self.add_opcodes(
            ASTORE_name('#value'),
            ALOAD_name('#module'),

            JavaOpcodes.GETFIELD('org/python/types/Module', '__dict__', 'Ljava/util/Map;'),

            ALOAD_name('#value'),
            java.Map.putAll(),

            free_name('#value')
        )

    def store_local(self):
        self.store_global()

    def load_name(self, name):
        self.add_opcodes(
            ALOAD_name('#module'),
            python.Object.get_attribute(name),
        )

    def load_globals(self):
        self.add_opcodes(
            ALOAD_name('#module'),
            JavaOpcodes.GETFIELD('org/python/types/Module', '__dict__', 'Ljava/util/Map;'),
        )

    def load_locals(self):
        self.load_globals()

    def load_vars(self):
        self.load_globals()

    def delete_name(self, name):
        self.add_opcodes(
            ALOAD_name('#module'),
            python.Object.del_attr(name),
        )

    def add_function(self, name, code, parameter_signatures, return_signature):
        if code.co_flags & CO_GENERATOR:
            # Generator method.
            function = GeneratorFunction(
                self,
                name=name,
                code=code,
                generator=code.co_name,
                parameters=parameter_signatures,
                returns=return_signature,
                static=True
            )
        else:
            # Normal function.
            function = Function(
                self,
                name=name,
                code=code,
                parameters=parameter_signatures,
                returns=return_signature,
                static=True,
            )

        # Add the function to the list that need to be
        # transpiled into Java functions
        self.functions.append(function)

        # Add a definition of the callable object
        self.add_callable(function)

        return function

    def add_class(self, class_name, extends, implements):
        from .klass import Class

        klass = Class(
            self,
            name=class_name,
            extends=extends,
            implements=implements,
        )

        self.classes.append(klass)

        self.add_opcodes(
            # Stack contains the bases list
            ASTORE_name('#bases'),

            # DEBUG("FORCE LOAD OF CLASS %s AT DEFINITION" % klass.descriptor),
            # - class
            JavaOpcodes.LDC_W(JavaClassref(klass.descriptor)),

            # - name
            JavaOpcodes.LDC_W(klass.name),

            # - bases
            ALOAD_name('#bases'),

            # - dict
            JavaOpcodes.ACONST_NULL(),

            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Type',
                'declarePythonType',
                args=[
                    'Ljava/lang/Class;',
                    'Ljava/lang/String;',
                    'Ljava/util/List;',
                    'Ljava/util/Map;'
                ],
                returns='Lorg/python/types/Type;'
            ),

            free_name('#bases')
        )

        self.store_name(klass.name)

        self.add_opcodes(
            JavaOpcodes.INVOKESTATIC(
                klass.descriptor,
                'class$init',
                args=[],
                returns='V'
            ),
        )

        return klass

    def visitor_teardown(self):
        self.add_opcodes(
            JavaOpcodes.RETURN(),
        )

    def transpile(self):
        """Convert a materialized Python code definition into a list of Java
        Classfile definitions.

        Returns a list of triples:
            (namespace, class_name, javaclassfile)

        The list contains the classfile for the module, plus and classes
        defined in the module.
        """
        # If there is any static content, generate a classfile
        # for this module
        classfile = JavaClass(self.class_descriptor, extends='org/python/types/Module')
        classfile.attributes.append(JavaSourceFile(os.path.basename(self.sourcefile)))

        # Add a static method to the module, populated with the
        # body content of the module.
        static_init = JavaMethod('module$import', '()V', public=False)
        static_init.attributes.append(self.transpile_code())
        classfile.methods.append(static_init)

        # Add a dummy init method.
        classfile.methods.append(
            JavaMethod(
                '<init>',
                '()V',
                attributes=[
                    JavaCode(
                        max_stack=1,
                        max_locals=1,
                        code=[
                            JavaOpcodes.ALOAD_0(),
                            JavaOpcodes.INVOKESPECIAL('org/python/types/Module', '<init>', args=[], returns='V'),
                            JavaOpcodes.RETURN(),
                        ],
                    ),
                ]
            ),
        )

        # Add a __new__ method to set the filename of the module.
        new_method = Accumulator({'self': 0, 'klass': 1})

        new_method.add_opcodes(
            # Call super.__new__()
            ALOAD_name('self'),
            ALOAD_name('klass'),
            JavaOpcodes.INVOKESPECIAL(
                'org/python/types/Module',
                '__new__',
                args=['Lorg/python/Object;'],
                returns='Lorg/python/Object;'
            ),

            ALOAD_name('self'),
            JavaOpcodes.GETFIELD(
                'org/python/types/Object',
                '__dict__',
                'Ljava/util/Map;'
            ),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__file__"),
            python.Str(self.sourcefile),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__package__"),
            python.Str(self.name),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__name__"),
            python.Str(self.full_name),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__builtins__"),
            python.NONE(),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__cached__"),
            python.NONE(),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__doc__"),
            python.NONE(),
            java.Map.put(),

            JavaOpcodes.DUP(),
            JavaOpcodes.LDC_W("__loader__"),
            python.Str(),
            java.Map.put(),
        )

        # Python 3.6 introduced __annotations__ at the module level.
        if sys.hexversion > 0x03060000:
            new_method.add_opcodes(
                JavaOpcodes.DUP(),
                JavaOpcodes.LDC_W("__annotations__"),
                python.Dict(),
                java.Map.put(),
            )

        new_method.add_opcodes(
            # last entry doesn't need to be duped, because we don't need to
            # reuse the value on the stack.
            JavaOpcodes.LDC_W("__spec__"),
            python.Str(),
            java.Map.put(),

            JavaOpcodes.ARETURN()
        )

        classfile.methods.append(
            JavaMethod(
                '__new__',
                '(Lorg/python/Object;)Lorg/python/Object;',
                attributes=[
                    JavaCode(
                        max_stack=new_method.max_stack(),
                        max_locals=new_method.max_locals(),
                        code=new_method.opcodes,
                    ),
                ]
            )
        )

        # Add any static methods defined in the module
        for function in self.functions:
            classfile.methods.extend(function.transpile())

        # The list of classfiles that will be returned will contain
        # at least one entry - the class for the module itself.
        classfiles = [(
            self.namespace,
            self.name + ('/__init__' if self.has_init_file else ''),
            classfile
        )]
        # Also output any classes defined in this module.
        for klass in self.classes:
            classfiles.append(klass.transpile())

        return classfiles



from ..utils import TranspileTestCase


class JavaTests(TranspileTestCase):

    def test_multiple_constructors(self):
        "The appropriate constructor for a native Java class can be interpolated from args"
        self.assertJavaExecution(
            """
            from java.lang import StringBuilder

            builder = StringBuilder("Hello, ")

            builder.append("world")

            print(builder)
            print("Done.")
            """,
            """
            Hello, world
            Done.
            """,
            run_in_function=False)

    def test_most_specific_constructor(self):
        "The most specific constructor for a native Java class will be selected based on argument."
        self.assertJavaExecution(
            """
            from com.example import MyClass

            obj1 = MyClass()
            obj2 = MyClass(1.234)
            obj3 = MyClass(3742)

            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public MyClass() {
                        System.out.println("No argument");
                    }

                    public MyClass(int arg) {
                        System.out.println("Integer argument " + arg);
                    }

                    public MyClass(double arg) {
                        System.out.println("Double argument " + arg);
                    }
                }
                """
            },
            out="""
            No argument
            Double argument 1.234
            Integer argument 3742
            Done.
            """, run_in_function=False)

    def test_field(self):
        "Native fields on an instance can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj1 = MyClass()
            print("Field is", MyClass.field)
            print("Field from instance is", obj1.field)
            obj1.field = 37
            print("Updated Field from instance is", obj1.field)
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public int field = 42;
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Field is <unbound native field public int com.example.MyClass.field>
                Field from instance is 42
                Updated Field from instance is 37
                Done.
                """)

    def test_static_field(self):
        "Class constants can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj1 = MyClass()
            print("Static field is", MyClass.static_field)
            MyClass.static_field = 37
            print("Updated static field is", MyClass.static_field)
            print("Static field from instance is", obj1.static_field)
            MyClass.static_field = 42
            print("Updated static field from instance is", obj1.static_field)
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public static int static_field = 42;
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Static field is 42
                Updated static field is 37
                Static field from instance is 37
                Updated static field from instance is 42
                Done.
                """)

    def test_superclass_field(self):
        "Native fields defined on a superclass can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyBase, MyClass

            print("Base class is", MyBase)
            print("Class is", MyClass)
            obj1 = MyClass()
            print("Base field on superclass is", MyBase.base_field)
            print("Base field is", MyClass.base_field)
            print("Base field from instance is", obj1.base_field)
            print("Field is", MyClass.field)
            print("Field from instance is", obj1.field)
            print("Done.")
            """,
            java={
                'com/example/MyBase': """
                package com.example;

                public class MyBase {
                    public int base_field = 37;
                }
                """,
                'com/example/MyClass': """
                package com.example;

                public class MyClass extends MyBase {
                    public int field = 42;
                }
                """
            },
            out="""
                Base class is <class 'com.example.MyBase'>
                Class is <class 'com.example.MyClass'>
                Base field on superclass is <unbound native field public int com.example.MyBase.base_field>
                Base field is <unbound native field public int com.example.MyBase.base_field>
                Base field from instance is 37
                Field is <unbound native field public int com.example.MyClass.field>
                Field from instance is 42
                Done.
                """)

    def test_superclass_static_field(self):
        "Native static fields defined on a superclass can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyBase, MyClass

            print("Base class is", MyBase)
            print("Class is", MyClass)
            obj1 = MyClass()
            print("Static base field on superclass is", MyBase.base_static_field)
            print("Static base field is", MyClass.base_static_field)
            print("Static base field from instance is", obj1.base_static_field)
            print("Static field is", MyClass.static_field)
            print("Static field from instance is", obj1.static_field)
            print("Done.")
            """,
            java={
                'com/example/MyBase': """
                package com.example;

                public class MyBase {
                    public static int base_static_field = 37;
                }
                """,
                'com/example/MyClass': """
                package com.example;

                public class MyClass extends MyBase {
                    public static int static_field = 42;
                }
                """
            },
            out="""
                Base class is <class 'com.example.MyBase'>
                Class is <class 'com.example.MyClass'>
                Static base field on superclass is 37
                Static base field is 37
                Static base field from instance is 37
                Static field is 42
                Static field from instance is 42
                Done.
                """)

    def test_constant(self):
        "Instance constants can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj1 = MyClass()
            print("Constant is", MyClass.CONSTANT)
            print("Constant from instance is", obj1.CONSTANT)
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public final int CONSTANT = 42;
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Constant is <unbound native field public final int com.example.MyClass.CONSTANT>
                Constant from instance is 42
                Done.
                """)

    def test_static_constant(self):
        "Class constants can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj1 = MyClass()
            print("Static constant is", MyClass.STATIC_CONSTANT)
            print("Static constant from instance is", obj1.STATIC_CONSTANT)
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public static final int STATIC_CONSTANT = 42;
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Static constant is 42
                Static constant from instance is 42
                Done.
                """)

    def test_method(self):
        "Native methods on an instance can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj = MyClass()
            print("Method is", MyClass.method)
            print("Method from instance is", obj.method)
            obj.method()
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public void method() {
                        System.out.println("Hello from the instance!");
                    }
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Method is <native function com.example.MyClass.method>
                Method from instance is <bound native method com.example.MyClass.method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the instance!
                Done.
                """) # noqa

    def test_static_method(self):
        "Native static methods on an instance can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            print("Class is", MyClass)
            obj = MyClass()
            print("Static method is", MyClass.method)
            MyClass.method()
            print("Static method from instance is", obj.method)
            obj.method()
            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public static void method() {
                        System.out.println("Hello from the class!");
                    }
                }
                """
            },
            out="""
                Class is <class 'com.example.MyClass'>
                Static method is <native function com.example.MyClass.method>
                Hello from the class!
                Static method from instance is <bound native method com.example.MyClass.method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the class!
                Done.
                """) # noqa

    def test_superclass_method(self):
        "Native methods defined on a superclass can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyBase, MyClass

            print("Base class is", MyBase)
            print("Class is", MyClass)

            print("Base method on superclass is", MyBase.base_method)
            print("Method on superclass is", MyBase.method)

            print("Base method is", MyClass.base_method)
            print("Method is", MyClass.method)

            obj1 = MyBase()
            print("Base method from superclass instance is", obj1.base_method)
            obj1.base_method()
            print("Method from superclass instance is", obj1.method)
            obj1.method()

            obj2 = MyClass()
            print("Base method from instance is", obj2.base_method)
            obj2.base_method()
            print("Method from instance is", obj2.method)
            obj2.method()
            print("Done.")
            """,
            java={
                'com/example/MyBase': """
                package com.example;

                public class MyBase {
                    public void base_method() {
                        System.out.println("Hello from the base!");
                    }

                    public void method() {
                        System.out.println("Goodbye from the base!");
                    }
                }
                """,
                'com/example/MyClass': """
                package com.example;

                public class MyClass extends MyBase {
                    public void method() {
                        System.out.println("Hello from the instance!");
                    }
                }
                """
            },
            out="""
                Base class is <class 'com.example.MyBase'>
                Class is <class 'com.example.MyClass'>
                Base method on superclass is <native function com.example.MyBase.base_method>
                Method on superclass is <native function com.example.MyBase.method>
                Base method is <native function com.example.MyBase.base_method>
                Method is <native function com.example.MyClass.method>
                Base method from superclass instance is <bound native method com.example.MyBase.base_method of <Native com.example.MyBase object at 0xXXXXXXXX>>
                Hello from the base!
                Method from superclass instance is <bound native method com.example.MyBase.method of <Native com.example.MyBase object at 0xXXXXXXXX>>
                Goodbye from the base!
                Base method from instance is <bound native method com.example.MyBase.base_method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the base!
                Method from instance is <bound native method com.example.MyClass.method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the instance!
                Done.
                """) # noqa

    def test_superclass_static_method(self):
        "Native static methods defined on a superclass can be accessed"
        self.assertJavaExecution(
            """
            from com.example import MyBase, MyClass

            print("Base class is", MyBase)
            print("Class is", MyClass)

            print("Static base method on superclass is", MyBase.base_static_method)
            MyBase.base_static_method()
            print("Static method on superclass is", MyBase.static_method)
            MyBase.static_method()

            print("Static base method is", MyClass.base_static_method)
            MyClass.base_static_method()
            print("Static method is", MyClass.static_method)
            MyClass.static_method()

            obj1 = MyBase()
            print("Base static method from superclass instance is", obj1.base_static_method)
            obj1.base_static_method()
            print("Static method from superclass instance is", obj1.static_method)
            obj1.static_method()

            obj2 = MyClass()
            print("Base static method from instance is", obj2.base_static_method)
            obj2.base_static_method()
            print("Static method from instance is", obj2.static_method)
            obj2.static_method()
            print("Done.")
            """,
            java={
                'com/example/MyBase': """
                package com.example;

                public class MyBase {
                    public static void base_static_method() {
                        System.out.println("Hello from the base!");
                    }

                    public static void static_method() {
                        System.out.println("Goodbye from the base!");
                    }
                }
                """,
                'com/example/MyClass': """
                package com.example;

                public class MyClass extends MyBase {
                    public static void static_method() {
                        System.out.println("Hello from the class!");
                    }
                }
                """
            },
            out="""
                Base class is <class 'com.example.MyBase'>
                Class is <class 'com.example.MyClass'>
                Static base method on superclass is <native function com.example.MyBase.base_static_method>
                Hello from the base!
                Static method on superclass is <native function com.example.MyBase.static_method>
                Goodbye from the base!
                Static base method is <native function com.example.MyBase.base_static_method>
                Hello from the base!
                Static method is <native function com.example.MyClass.static_method>
                Hello from the class!
                Base static method from superclass instance is <bound native method com.example.MyBase.base_static_method of <Native com.example.MyBase object at 0xXXXXXXXX>>
                Hello from the base!
                Static method from superclass instance is <bound native method com.example.MyBase.static_method of <Native com.example.MyBase object at 0xXXXXXXXX>>
                Goodbye from the base!
                Base static method from instance is <bound native method com.example.MyBase.base_static_method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the base!
                Static method from instance is <bound native method com.example.MyClass.static_method of <Native com.example.MyClass object at 0xXXXXXXXX>>
                Hello from the class!
                Done.
                """)  # noqa

    def test_inner_class_constant(self):
        "Constants on an inner class can be accessed"
        self.assertJavaExecution(
            """
            from com.example import OuterClass

            print("Outer class is", OuterClass)
            print("Outer constant is", OuterClass.OUTER_CONSTANT)

            print("Inner class is", OuterClass.InnerClass)
            print("Inner constant is", OuterClass.InnerClass.INNER_CONSTANT)

            print("Done.")
            """,
            java={
                'com/example/OuterClass': """
                package com.example;

                public class OuterClass {
                    public static final int OUTER_CONSTANT = 42;

                    public static class InnerClass {
                        public static final int INNER_CONSTANT = 37;
                    }
                }
                """
            },
            out="""
                Outer class is <class 'com.example.OuterClass'>
                Outer constant is 42
                Inner class is <class 'com.example.OuterClass$InnerClass'>
                Inner constant is 37
                Done.
                """)

    def test_inner_class_method(self):
        "Inner classes can be instantiated, and methods invoked"
        self.assertJavaExecution(
            """
            from com.example import OuterClass

            print("Outer class is", OuterClass)
            obj1 = OuterClass()
            obj1.method()

            print("Inner class is", OuterClass.InnerClass)
            obj2 = OuterClass.InnerClass(obj1)
            obj2.method()

            print("Done.")
            """,
            java={
                'com/example/OuterClass': """
                package com.example;

                public class OuterClass {
                    public class InnerClass {
                        public void method() {
                            System.out.println("Hello from the inside!");
                        }
                    }

                    public void method() {
                        System.out.println("Hello from the outside!");
                    }
                }
                """
            },
            out="""
                Outer class is <class 'com.example.OuterClass'>
                Hello from the outside!
                Inner class is <class 'com.example.OuterClass$InnerClass'>
                Hello from the inside!
                Done.
                """)

    def test_static_inner_class_constant(self):
        "Constants on a static inner class can be accessed"
        self.assertJavaExecution(
            """
            from com.example import OuterClass

            print("Outer class is", OuterClass)
            print("Outer constant is", OuterClass.OUTER_CONSTANT)

            print("Inner class is", OuterClass.InnerClass)
            print("Inner constant is", OuterClass.InnerClass.INNER_CONSTANT)

            print("Done.")
            """,
            java={
                'com/example/OuterClass': """
                package com.example;

                public class OuterClass {
                    public static final int OUTER_CONSTANT = 42;

                    public static class InnerClass {
                        public static final int INNER_CONSTANT = 37;
                    }
                }
                """
            },
            out="""
                Outer class is <class 'com.example.OuterClass'>
                Outer constant is 42
                Inner class is <class 'com.example.OuterClass$InnerClass'>
                Inner constant is 37
                Done.
                """)

    def test_static_inner_class_method(self):
        "Static inner classes can be instantiated, and methods invoked"
        self.assertJavaExecution(
            """
            from com.example import OuterClass

            print("Outer class is", OuterClass)
            obj1 = OuterClass()
            obj1.method()

            print("Inner class is", OuterClass.InnerClass)
            obj2 = OuterClass.InnerClass()
            obj2.method()

            print("Done.")
            """,
            java={
                'com/example/OuterClass': """
                package com.example;

                public class OuterClass {
                    public static class InnerClass {
                        public void method() {
                            System.out.println("Hello from the inside!");
                        }
                    }

                    public void method() {
                        System.out.println("Hello from the outside!");
                    }
                }
                """
            },
            out="""
                Outer class is <class 'com.example.OuterClass'>
                Hello from the outside!
                Inner class is <class 'com.example.OuterClass$InnerClass'>
                Hello from the inside!
                Done.
                """)

    def test_primitive_conversion(self):
        "Primitive types are converted correctly"
        self.assertJavaExecution(
            """
            from com.example import MyObject

            obj = MyObject()

            result = obj.method(3)
            print('Result is', result)

            result = obj.method(3.14159)
            print('Result is', result)

            result = obj.method(True)
            print('Result is', result)

            print("Done.")
            """,
            java={
                'com/example/MyObject': """
                package com.example;

                public class MyObject {
                    public int method(int x) {
                        System.out.println("Hello from int method: " + x);
                        return x * 2;
                    }

                    public float method(float x) {
                        System.out.println("Hello from float method: " + x);
                        return x * 2.5f;
                    }

                    public boolean method(boolean x) {
                        System.out.println("Hello from boolean method: " + x);
                        return !x;
                    }
                }
                """
            },
            out="""
                Hello from int method: 3
                Result is 6
                Hello from float method: 3.14159
                Result is 7.85397...
                Hello from boolean method: true
                Result is False
                Done.
                """)

    def test_primitive_zero_conversion(self):
        "Primitive representations of 'false' values are converted correctly"
        self.assertJavaExecution(
            """
            from com.example import MyObject

            obj = MyObject()

            result = obj.method(0)
            print('Result is', result)

            result = obj.method(0.0)
            print('Result is', result)

            result = obj.method(False)
            print('Result is', result)

            print("Done.")
            """,
            java={
                'com/example/MyObject': """
                package com.example;

                public class MyObject {
                    public int method(int x) {
                        System.out.println("Hello from int method: " + x);
                        return x * 2;
                    }

                    public float method(float x) {
                        System.out.println("Hello from float method: " + x);
                        return x * 2.5f;
                    }

                    public boolean method(boolean x) {
                        System.out.println("Hello from boolean method: " + x);
                        return !x;
                    }
                }
                """
            },
            out="""
                Hello from int method: 0
                Result is 0
                Hello from float method: 0.0
                Result is 0.0
                Hello from boolean method: false
                Result is True
                Done.
                """)

    def test_attribute_on_reference(self):
        self.assertJavaExecution(
            """
            from com.example import BaseObject, Controller

            class TestObject(extends=com.example.BaseObject):
                def __init__(self, v1, v2):
                    # super().__init__(v1)
                    self.extra1 = v2

                def process(self) -> None:
                    print("Processed Base data is %s" % self.base)
                    print("Processed Extra data 1 is %s" % self.extra1)
                    print("Processed Extra data 2 is %s" % self.extra2)

            obj = TestObject(37, 99)
            obj.extra2 = 42

            print("Base data is %s" % obj.base)
            print("Extra data 1 is %s" % obj.extra1)
            print("Extra data 2 is %s" % obj.extra2)

            controller = Controller()

            controller.setObject(obj)

            controller.poke()

            print("Done.")
            """,
            java={
                'com/example/ObjIFace': """
                package com.example;

                public interface ObjIFace {
                    public void process();
                }

                """,
                'com/example/Controller': """
                package com.example;

                public class Controller {
                    public ObjIFace obj;

                    public Controller() {
                    }

                    public void setObject(ObjIFace o) {
                        this.obj = o;
                    }

                    public void poke() {
                        this.obj.process();
                    }
                }

                """,
                'com/example/BaseObject': """
                package com.example;

                public class BaseObject implements com.example.ObjIFace {
                    public int base;

                    public BaseObject() {
                        base = 1;
                    }

                    public BaseObject(int v) {
                        base = v;
                    }

                    public void process() {
                        System.out.println("Do nothing...");
                    }
                }
                """
            },
            out="""
                Base data is 1
                Extra data 1 is 99
                Extra data 2 is 42
                Processed Base data is 1
                Processed Extra data 1 is 99
                Processed Extra data 2 is 42
                Done.
                """,
            run_in_function=False)



import unittest

from .utils import adjust, cleanse_java, cleanse_python, TranspileTestCase


class AdjustTests(unittest.TestCase):
    def assertEqualOutput(self, actual, expected):
        self.assertEqual(adjust(actual), adjust(expected))

    def test_adjust(self):
        "Test input can be stripped of leading spaces."
        self.assertEqual("""for i in range(0, 10):
    print('hello, world')
print('Done.')
""", adjust("""
            for i in range(0, 10):
                print('hello, world')
            print('Done.')
        """))

    def test_adjust_no_leading_space(self):
        self.assertEqual("""for i in range(0, 10):
    print('hello, world')
print('Done.')
""", adjust("""for i in range(0, 10):
    print('hello, world')
print('Done.')
"""))


class AssertCodeExecutionTests(TranspileTestCase):
    @unittest.expectedFailure
    def test_fail_unexpected_exception(self):
        self.assertCodeExecution("""
            raise ValueError
            """)

    def test_allow_expected_exception(self):
        self.assertCodeExecution("""
            raise ValueError
            """, exits_early=True)

    @unittest.expectedFailure
    def test_fail_missing_expected_exception(self):
        self.assertCodeExecution("""
            print('Done.')
            """, exits_early=True)


class JavaNormalizationTests(unittest.TestCase):
    def assertNormalized(self, actual, expected):
        self.assertEqual(cleanse_java(adjust(actual), None), adjust(expected))

    def test_no_exception(self):
        self.assertNormalized(
            """
            Hello, world.
            """,
            """
            Hello, world.
            """
        )

    def test_exception_in_clinit(self):
        self.assertNormalized(
            """
            Exception in thread "main" java.lang.ExceptionInInitializerError
            Caused by: IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.<clinit>(test.py:2)
            """,
            """
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:2
            """
        )

    def test_exception_in_module_init(self):
        self.assertNormalized(
            """
            Exception in thread "main" NameError: name 'y' is not defined
                at org.python.types.Module.__getattribute__(Module.java:32)
                at python.example.__init__.module$import(example.py:2)
                at python.example.__init__.main(example.py)
            """,
            """
            ### EXCEPTION ###
            NameError: name 'y' is not defined
                example.py:2
            """
        )

    def test_exception_in_clinit_after_output(self):
        self.assertNormalized(
            """
            Hello, world.
            Exception in thread "main" java.lang.ExceptionInInitializerError
            Caused by: IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.<clinit>(test.py:2)
            """,
            """
            Hello, world.
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:2
            """
        )

    def test_exception_in_clinit_after_output_windows(self):
        self.assertNormalized(
            """
            Hello, world.
            java.lang.ExceptionInInitializerError
            Caused by: IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.<clinit>(test.py:2)
            """,
            """
            Hello, world.
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:2
            """
        )

    def test_exception_in_clinit_after_output_testdaemon(self):
        self.assertNormalized(
            """
            False
            True
            Exception in thread "main" java.lang.ExceptionInInitializerError
            \tat sun.reflect.NativeMethodAccessorImpl.invoke0(Native Method)
            \tat sun.reflect.NativeMethodAccessorImpl.invoke(NativeMethodAccessorImpl.java:62)
            \tat sun.reflect.DelegatingMethodAccessorImpl.invoke(DelegatingMethodAccessorImpl.java:43)
            \tat java.lang.reflect.Method.invoke(Method.java:497)
            \tat python.testdaemon.TestDaemon.main(TestDaemon.java:66)
            Caused by: KeyError: 'c'
                at org.python.types.Dict.__getitem__(Dict.java:142)
                at python.test.__init__.<clinit>(test.py:4)
                ... 5 more
            """,
            """
            False
            True
            ### EXCEPTION ###
            KeyError: 'c'
                test.py:4
            """
        )

    def test_exception_in_method(self):
        self.assertNormalized(
            """
            Exception in thread "main" IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.main(test.py:3)
            """,
            """
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:3
            """
        )

    def test_exception_in_method_windows(self):
        self.assertNormalized(
            """
            IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.main(test.py:3)
            """,
            """
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:3
            """
        )

    def test_exception_in_method_after_output(self):
        self.assertNormalized(
            """
            Hello, world.
            Exception in thread "main" IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.main(test.py:3)
            """,
            """
            Hello, world.
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:3
            """
        )

    def test_exception_in_method_after_output_windows(self):
        self.assertNormalized(
            """
            Hello, world.
            IndexError: list index out of range
                at org.python.types.List.__getitem__(List.java:100)
                at org.python.types.List.__getitem__(List.java:85)
                at python.test.main(test.py:3)
            """,
            """
            Hello, world.
            ### EXCEPTION ###
            IndexError: list index out of range
                test.py:3
            """
        )

    def test_exception_in_constructor(self):
        self.assertNormalized(
            """
            Exception in thread "main" java.lang.ExceptionInInitializerError
            Caused by: UnboundLocalError: local variable 'x' referenced before assignment
              at python.example.Foo.__init__(example.py:44)
              at sun.reflect.NativeMethodAccessorImpl.invoke0(Native Method)
              at sun.reflect.NativeMethodAccessorImpl.invoke(NativeMethodAccessorImpl.java:57)
              at sun.reflect.DelegatingMethodAccessorImpl.invoke(DelegatingMethodAccessorImpl.java:43)
              at java.lang.reflect.Method.invoke(Method.java:606)
              at org.python.types.Method.invoke(Method.java:66)
              at python.example.Foo.<init>(example.py)
              at sun.reflect.NativeConstructorAccessorImpl.newInstance0(Native Method)
              at sun.reflect.NativeConstructorAccessorImpl.newInstance(NativeConstructorAccessorImpl.java:57)
              at sun.reflect.DelegatingConstructorAccessorImpl.newInstance(DelegatingConstructorAccessorImpl.java:45)
              at java.lang.reflect.Constructor.newInstance(Constructor.java:526)
              at org.python.types.Constructor.invoke(Constructor.java:25)
              at python.example.<clinit>(example.py:51)
            """,
            """
            ### EXCEPTION ###
            UnboundLocalError: local variable 'x' referenced before assignment
                example.py:51
                example.py:44
            """
        )

    def test_memory_reference(self):
        self.assertNormalized(
            """
            Class is <class 'com.example.MyClass'>
            Method is <native function com.example.MyClass.method>
            Method from instance is <bound native method com.example.MyClass.method of <Native class com.example.MyClass object at 0x1eb19f4e>>
            Hello from the instance!
            Done.
            """,  # noqa
            """
            Class is <class 'com.example.MyClass'>
            Method is <native function com.example.MyClass.method>
            Method from instance is <bound native method com.example.MyClass.method of <Native class com.example.MyClass object at 0xXXXXXXXX>>
            Hello from the instance!
            Done.
            """  # noqa
        )


class PythonNormalizationTests(unittest.TestCase):
    def assertNormalized(self, actual, expected):
        self.assertEqual(cleanse_python(adjust(actual), None), adjust(expected))

    def test_no_exception(self):
        self.assertNormalized(
            """
            Hello, world.
            """,
            """
            Hello, world.
            """
        )

    def test_exception(self):
        self.assertNormalized(
            """
            Traceback (most recent call last):
              File "test.py", line 3, in <module>
                print(x & y)
            TypeError: unsupported operand type(s) for &: 'float' and 'bool'
            """,
            """
            ### EXCEPTION ###
            TypeError: unsupported operand type(s) for &: 'float' and 'bool'
                test.py:3
            """
        )

    def test_exception_with_other_text(self):
        self.assertNormalized(
            """
            Hello, world.
            Traceback (most recent call last):
              File "test.py", line 3, in <module>
                print(x & y)
            TypeError: unsupported operand type(s) for &: 'float' and 'bool'
            """,
            """
            Hello, world.
            ### EXCEPTION ###
            TypeError: unsupported operand type(s) for &: 'float' and 'bool'
                test.py:3
            """
        )

    def test_memory_reference(self):
        self.assertNormalized(
            """
            Class is <class 'com.example.MyClass'>
            Method is <native function com.example.MyClass.method>
            Method from instance is <bound native method com.example.MyClass.method of <Native class com.example.MyClass object at 0x1eb19f4e>>
            Hello from the instance!
            Done.
            """,  # noqa
            """
            Class is <class 'com.example.MyClass'>
            Method is <native function com.example.MyClass.method>
            Method from instance is <bound native method com.example.MyClass.method of <Native class com.example.MyClass object at 0xXXXXXXXX>>
            Hello from the instance!
            Done.
            """  # noqa
        )


class JavaBootstrapTests(TranspileTestCase):
    def test_java_code(self):
        "You can supply Java code and use it from within Python"
        self.assertJavaExecution(
            """
            from com.example import MyClass

            obj = MyClass()

            obj.doStuff()

            print("Done.")
            """,
            java={
                'com/example/MyClass': """
                package com.example;

                public class MyClass {
                    public void doStuff() {
                        System.out.println("Hello from Java");
                    }
                }
                """
            },
            out="""
            Hello from Java
            Done.
            """,
        )



from .. utils import TranspileTestCase, BuiltinFunctionTestCase, BuiltinTwoargFunctionTestCase


class RoundTests(TranspileTestCase):
    pass


class BuiltinRoundFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["round"]


class BuiltinRoundTwoargFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["round"]
    python_types = [
        'None',
        'NotImplemented',
        'bool',
        'bytearray',
        'bytes',
        'class',
        'complex',
        'dict',
        'float',
        'frozenset',
        'int',
        'list',
        'range',
        'set',
        'slice',
        'str',
        'tuple',
        'obj',
    ]

    not_implemented_types = [
        'test_float',
    ]

    not_implemented = [
        'test_bool_bool',
        'test_bool_int',
        'test_int_bool',
        'test_int_int',
        'test_obj_class',
    ]

    not_implemented_versions = {
        'test_bool_None': (3.5, 3.6),
        'test_int_None': (3.5, 3.6),
    }

    for not_implemented_type in not_implemented_types:
        for python_type in python_types:
            not_implemented.append('_'.join([not_implemented_type, python_type]))



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class OctTests(TranspileTestCase):
    pass


class BuiltinOctFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["oct"]

    not_implemented = [
    ]



import sys
from os import path
sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from utils import *

def test_small_integers(test_case):
    print("Running", "test_small_integers")
    test_case.runAsJava(adjust("""
        for i in range(100):
            for j in range(100):
                x = -5
                for k in range(-5, 257):
                    x = x + 1
        """), timed=True)

def test_booleans(test_case):
    print("Running", "test_booleans")
    test_case.runAsJava(adjust("""
        for i in range(100000):
            x = True
            y = False

            x = (not x and y)
            y = x != y

            x = True and True and True
            y = x or False

            x = not True or (not x) or True
            y = y == y

            x = y != (not x)
            y = False or True or (not y)

            x = (x != x) or (not False)
            y = y and y and (not x)

            x = False or False or False or False
            y = True and True and True and True

            x = 1 != 0
            y = (not (not (not (not (not (not (not x)))))))

            x = ((not y) and y) == ((not (not x)) or y or (not y))
            y = False or False or (not True) or (not True) or (not True)

            x = True and (not False) and True and (not False) and False
            y = (False != False) != False
        """), timed=True)

def test_global_var_load(test_case):
    print("Running", "test_global_var_load")
    test_case.runAsJava(adjust("""
        x = 1
        y = 2

        for i in range(100000):
            print(x)
            print(y)
            print(x+y)
            print(x-y)
            print(x)
            print(y)
            print(x*y)
            print(x/y)
            print(x)
            print(y)
    """), timed=True)

def test_class_var_load(test_case):
    print("Running", "test_class_var_load")
    test_case.runAsJava(adjust("""
        class Animal:
            def __init__(self, name, sound):
                self.name = name
                self.sound = sound

            def speak(self):
                print(self.name)
                print(self.sound)
                print(self.name)
                print(self.sound)
                print(self.name)
                print(self.sound)

        lapo = Animal("Lapo", "Bow Wow")
        for i in range(100000):
            lapo.speak()
            lapo.speak()
            lapo.speak()

    """), timed=True)

def test_function_var_load(test_case):
    print("Running", "test_function_var_load")
    test_case.runAsJava(adjust("""
        def foo():
            x = 1
            y = 2
            print(x)
            print(y)
            print(x+y)
            print(x-y)
            print(x)
            print(y)
            print(x*y)
            print(x/y)
            print(x)
            print(y)

        for i in range(100000):
            foo()
    """), timed=True)

def test_code(test_case):
    print("Running", "test_code")
    test_case.runAsJava(adjust("""
        def main(n):
            def foo(n):
                return n + 1
            def bar(n):
                return n + 1
            def baz(n):
                return n + 1
            def buzz(n):
                return n + 1
            def bizz(n):
                return n + 1
            def bozz(n):
                return n + 1
            def yam(n):
                return n + 1
            def yarn(n):
                return n + 1
            return foo(bar(baz(buzz(bizz(bozz(yam(yarn(n))))))))
        for i in range(100000):
            main(i)
    """), timed=True)

def test_cmp(test_case):
    print("Running", "test_cmp")
    test_case.runAsJava(adjust("""
        x = None
        s = "mary had a little lamb"
        t = "humpty dumpty sat on a wall"

        for i in range(1000):
            for j in range(1000):
                x = s < t
                x = s <= t
                x = s == t
                x = s != t
                x = s > t
                x = s >= t

                x = 3 < 5
                x = 3 <= 5
                x = 3 == 5
                x = 3 != 5
                x = 3 > 5
                x = 3 >= 5

                x = 3 < True
                x = 3.0 <= 5
                x = None == 5
                x = [3] != 5.0
                x = [3] > [5]
                x = [3.0] > [5.0]
    """), timed=True)

def test_loops(test_case):
    print("Running", "test_loops")
    test_case.runAsJava(adjust("""
        for x in range(100):
            for y in range(100):
                for z in range(100):
                    for a in range(100):
                        pass
    """), timed=True)

def test_dict_get(test_case):
    print("Running", "test_dictionary_get")
    test_case.runAsJava(adjust("""
        dict = {1 : 2, "a" : "b"}
        for i in range(1000000):
            dict.get(i)
            dict.get(1)
            dict.get("a")
            dict.get(i)
            dict.get(1)
            dict.get("a")
    """), timed=True)

def test_dict_set(test_case):
    print("Running", "test_dictionary_set")
    test_case.runAsJava(adjust("""
        dict = {}
        for i in range(1000000):
            dict["moo"] = 1
            dict["quack"] = 2
            dict["woof"] = 3
            dict["meow"] = 4
            dict["cockadoodledoo"] = 5
            dict["hiss"] = 6
    """, timed=True)

def test_class_init(test_case):
    print("Running" , "test_class_init")
    test_case.runAsJava(adjust("""
        class A: pass
        class B: pass
        class C: pass
        class D: pass
        class E: pass
        class F: pass
        class G: pass
        class H: pass
        class I: pass
        class J: pass
        class K: pass
        class L: pass
        class M: pass
        class N: pass
        class O: pass
        class P: pass
        class Q: pass
        class R: pass
        class S: pass
        class T: pass
        class U: pass
        class V: pass
        class W: pass
        class X: pass
        class Y: pass
        class Z: pass

        a = "a"
        b = 1
        c = [1]
        d = {1 : 1}
        e = list([1])
        f = True
        g = 3.0
    """), timed=True)

def test_method(test_case):
    print("Running, test_method")
    test_case.runAsJava(adjust("""
        class MyClass:

            def A(self):
                print("A!")

            def B(self):
                print("B!")

            def C(self):
                print("C!")

        obj = MyClass()

        for i in range(1000000):
            obj.A()
            obj.B()
            obj.C()
    """), timed=True)

def main():
    test_case = TranspileTestCase()
    test_case.setUpClass()
    test_small_integers(test_case)
    test_booleans(test_case)
    test_global_var_load(test_case)
    test_class_var_load(test_case)
    test_function_var_load(test_case)
    test_code(test_case)
    test_loops(test_case)
    test_cmp(test_case)
    test_dict_get(test_case)
    test_dict_set(test_case)
    test_class_init(test_case)
    test_method(test_case)

if __name__== "__main__":
  main()



import codecs

from . import mutf_8
from .attributes import *  # noqa
from .constants import *  # noqa
from .fields import Field  # noqa
from .klass import Class  # noqa
from .methods import Method  # noqa
from .opcodes import *  # noqa

# Register the MUTF-8 codec
codecs.register(mutf_8.search_function)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class IntTests(TranspileTestCase):
    pass


class BuiltinIntFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["int"]

    not_implemented = [
        'test_bytearray',
        'test_bytes',
    ]



from unittest import expectedFailure

from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class ObjectTests(TranspileTestCase):
    @expectedFailure
    def test_setitem(self):
        self.assertCodeExecution("""
            class MyClass:
                def __setitem__(self, key, item):
                    print("In __setitem__")

            obj = MyClass()
            obj[0] = 0
            """)

    def test_getitem(self):
        self.assertCodeExecution("""
            class MyClass:
                def __getitem__(self, key):
                    print("In __getitem__")

            obj = MyClass()
            print(obj[0])
            """)

    @expectedFailure
    def test_getattribute(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            try:
                print(obj.a)
            except AttributeError as e:
                print(e)

            class MyClass1:
                def __getattribute__(self, name):
                    return name

            obj = MyClass1()
            print(obj.a)
        """)

    @expectedFailure
    def test_getattr(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            try:
                print(obj.a)
            except AttributeError as e:
                print(e)

            class MyClass1:
                def __getattr__(self, name):
                    return name

            obj = MyClass1()
            print(obj.a)
        """)

    @expectedFailure
    def test_hash(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            hash_code = obj.hash()
            print(isinstance(hash_code, int))
        """)

    @expectedFailure
    def test_repr(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            print(obj.repr())

            class MyClass1:
                def __repr__(self):
                    return "I am a MyClass1!"

            obj = MyClass1()
            print(obj.repr())
        """)

    def test_bytes(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            try:
                print(bytes(obj))
            except TypeError as e:
                print(e)
        """)

    def test_eq(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj1 = MyClass()
            obj2 = MyClass()
            print(obj1 == obj2)
            print(obj1 is obj2)

            class MyClass1:
                def __eq__(self, other):
                    return True

            obj1 = MyClass1()
            obj2 = MyClass1()
            print(obj1 == obj2)
            print(obj1 is obj2)
        """)

    def test_le(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj1 = MyClass()
            obj2 = MyClass()
            try:
                print(obj1 <= obj2)
            except TypeError as e:
                print(e)

            class MyClass1:
                def __le__(self, other):
                    return True

            obj1 = MyClass1()
            obj2 = MyClass1()
            print(obj1 <= obj2)
        """)

    def test_lt(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj1 = MyClass()
            obj2 = MyClass()
            try:
                print(obj1 < obj2)
            except TypeError as e:
                print(e)

            class MyClass1:
                def __lt__(self, other):
                    return True

            obj1 = MyClass1()
            obj2 = MyClass1()
            print(obj1 < obj2)
        """)

    def test_ge(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj1 = MyClass()
            obj2 = MyClass()
            try:
                print(obj1 >= obj2)
            except TypeError as e:
                print(e)

            class MyClass1:
                def __ge__(self, other):
                    return True

            obj1 = MyClass1()
            obj2 = MyClass1()
            print(obj1 >= obj2)
        """)

    def test_gt(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj1 = MyClass()
            obj2 = MyClass()
            try:
                print(obj1 > obj2)
            except TypeError as e:
                print(e)

            class MyClass1:
                def __gt__(self, other):
                    return True

            obj1 = MyClass1()
            obj2 = MyClass1()
            print(obj1 > obj2)
        """)


class UnaryObjectOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'obj'

    not_implemented = [
        'test_unary_not'
    ]


class BinaryObjectOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'obj'

    not_implemented = [
        'test_xor_class',
        'test_true_divide_class',
        'test_subtract_class',
        'test_subscr_class',
        'test_lshift_class',
        'test_rshift_class',
        'test_power_class',
        'test_or_class',
        'test_multiply_class',
        'test_modulo_class',
        'test_lt_class',
        'test_gt_class',
        'test_ge_class',
        'test_le_class',
        'test_add_class',
        'test_and_class',
        'test_floor_divide_class'
    ]

    is_flakey = [
        'test_eq_class',
        'test_ne_class',
    ]



class InplaceObjectOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'obj'

    not_implemented = [
        'test_multiply_tuple',
        'test_multiply_str',
        'test_multiply_list',
        'test_multiply_bytes',
        'test_multiply_bytearray',

        'test_xor_class',
        'test_true_divide_class',
        'test_subtract_class',
        'test_rshift_class',
        'test_power_class',
        'test_or_class',
        'test_multiply_class',
        'test_modulo_class',
        'test_lshift_class',
        'test_floor_divide_class',
        'test_and_class',
        'test_add_class',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class HexTests(TranspileTestCase):
    pass


class BuiltinHexFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["hex"]

    not_implemented = [
    ]



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class BytesTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = b'hello, world'
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_islower(self):
        self.assertCodeExecution("""
            print(b'abc'.islower())
            print(b''.islower())
            print(b'Abccc'.islower())
            print(b'HELLO WORD'.islower())
            print(b'@#$%!'.islower())
            print(b'hello world'.islower())
            print(b'hello world   '.islower())
            # TODO: uncomment when adding support for literal hex bytes
            #print(b'\xf0'.islower())

        """)
        # self.assertCodeExecution("""""")

    def test_isupper(self):
        self.assertCodeExecution("""
            print(b'abc'.isupper())
            print(b''.isupper())
            print(b'Abccc'.isupper())
            print(b'HELLO WORD'.isupper())
            print(b'@#$%!'.isupper())
            print(b'hello world'.isupper())
            print(b'hello world   '.isupper())
        """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = b'hello, world'
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_capitalize(self):
        self.assertCodeExecution(r"""
            print(b'hello, world'.capitalize())
            print(b'helloWORLD'.capitalize())
            print(b'HELLO WORLD'.capitalize())
            print(b'2015638687'.capitalize())
            print(b'\xc8'.capitalize())
        """)

    def test_partition(self):
        self.assertCodeExecution(r"""
            print(b'hello, world'.partition(b','))
            print(b'hello, world'.partition(b'h'))
            print(b'hello, world'.partition(b'd'))
            print(b'hello, world'.partition(b', '))
            print(b'hello, world'.partition(b'l'))
            print(b'2015638687'.partition(b'a'))
            print(b'\xc8'.partition(b' '))
        """)
        # self.assertCodeExecution(r"""
        #     print(b'hello, world'.partition(None))
        # """, exits_early=True)

    def test_repr(self):
        self.assertCodeExecution(r"""
            print(repr(b'\xc8'))
            print(repr(b'abcdef \xc8 abcdef'))
            print(repr(b'abcdef \xc8 abcdef\n\r\t'))
            print(b'abcdef \xc8 abcdef\n\r\t')
            for b in range(0, 256, 16):
                print(repr(bytes(range(b, b+16))))
            for b in range(0, 256, 16):
                print(bytes(range(b, b+16)))
        """)

    def test_iter(self):
        self.assertCodeExecution("""
            print([b for b in b''])
            print([b for b in b'hello world'])
        """)

    def test_getitem(self):

        self.assertCodeExecution("""
            x = b'0123456789'
            print("x[0] = ", x[0])
            print("x[-10] = ", x[-10])
            print("x[9] = ", x[9])
            #Start/Stop Empty w/Step +/-
            print("x[:] = ", x[:])
            print("x[::1] = ", x[::1])
            print("x[::2] = ", x[::2])
            print("x[::3] = ", x[::3])
            print("x[::-1] = ", x[::-1])
            print("x[::-2] = ", x[::-2])
            print("x[::-3] = ", x[::-3])
            #Empty Start Tests With Stop Bounds checks
            print("x[:9:1] = ", x[:9:1])
            print("x[:10:1] = ", x[:10:1])
            print("x[:11:1] = ", x[:11:1])
            print("x[:-9:1] = ", x[:-9:1])
            print("x[:-10:1] = ", x[:-10:1])
            print("x[:-11:1] = ", x[:-11:1])
            print("x[:9:2] = ", x[:9:2])
            print("x[:10:2] = ", x[:10:2])
            print("x[:11:2] = ", x[:11:2])
            print("x[:-9:2] = ", x[:-9:2])
            print("x[:-10:2] = ", x[:-10:2])
            print("x[:-11:2] = ", x[:-11:2])
            print("x[:9:3] = ", x[:9:3])
            print("x[:10:3] = ", x[:10:3])
            print("x[:11:3] = ", x[:11:3])
            print("x[:-9:3] = ", x[:-9:3])
            print("x[:-10:3] = ", x[:-10:3])
            print("x[:-11:3] = ", x[:-11:3])
            print("x[:9:-1] = ", x[:9:-1])
            print("x[:10:-1] = ", x[:10:-1])
            print("x[:11:-1] = ", x[:11:-1])
            print("x[:-9:-1] = ", x[:-9:-1])
            print("x[:-10:-1] = ", x[:-10:-1])
            print("x[:-11:-1] = ", x[:-11:-1])
            print("x[:9:-2] = ", x[:9:-2])
            print("x[:10:-2] = ", x[:10:-2])
            print("x[:11:-2] = ", x[:11:-2])
            print("x[:-9:-2] = ", x[:-9:-2])
            print("x[:-10:-2] = ", x[:-10:-2])
            print("x[:-11:-2] = ", x[:-11:-2])
            print("x[:9:-3] = ", x[:9:-3])
            print("x[:10:-3] = ", x[:10:-3])
            print("x[:11:-3] = ", x[:11:-3])
            print("x[:-9:-3] = ", x[:-9:-3])
            print("x[:-10:-3] = ", x[:-10:-3])
            print("x[:-11:-3] = ", x[:-11:-3])
            #Empty stop tests with stop bounds checks
            print("x[9::1] = ", x[9::1])
            print("x[10::1] = ", x[10::1])
            print("x[11::1] = ", x[11::1])
            print("x[-9::1] = ", x[-9::1])
            print("x[-10::1] = ", x[-10::1])
            print("x[-11::1] = ", x[-11::1])
            print("x[9::2] = ", x[9::2])
            print("x[10::2] = ", x[10::2])
            print("x[11::2] = ", x[11::2])
            print("x[-9::2] = ", x[-9::2])
            print("x[-10::2] = ", x[-10::2])
            print("x[-11::2] = ", x[-11::2])
            print("x[9::3] = ", x[9::3])
            print("x[10::3] = ", x[10::3])
            print("x[11::3] = ", x[11::3])
            print("x[-9::3] = ", x[-9::3])
            print("x[-10::3] = ", x[-10::3])
            print("x[-11::3] = ", x[-11::3])
            print("x[9::-1] = ", x[9::-1])
            print("x[10::-1] = ", x[10::-1])
            print("x[11::-1] = ", x[11::-1])
            print("x[-9::-1] = ", x[-9::-1])
            print("x[-10::-1] = ", x[-10::-1])
            print("x[-11::-1] = ", x[-11::-1])
            print("x[9::-2] = ", x[9::-2])
            print("x[10::-2] = ", x[10::-2])
            print("x[11::-2] = ", x[11::-2])
            print("x[-9::-2] = ", x[-9::-2])
            print("x[-10::-2] = ", x[-10::-2])
            print("x[-11::-2] = ", x[-11::-2])
            print("x[9::-3] = ", x[9::-3])
            print("x[10::-3] = ", x[10::-3])
            print("x[11::-3] = ", x[11::-3])
            print("x[-9::-3] = ", x[-9::-3])
            print("x[-10::-3] = ", x[-10::-3])
            print("x[-11::-3] = ", x[-11::-3])
            #other tests
            print("x[-5:] = ", x[-5:])
            print("x[:-5] = ", x[:-5])
            print("x[-2:-8] = ", x[-2:-8])
            print("x[100::-1] = ", x[100::-1])
            print("x[100:-100:-1] = ", x[100:-100:-1])
            print("x[:-100:-1] = ", x[:-100:-1])
            print("x[::-1] = ", x[::-1])
            print("x[::-2] = ", x[::-2])
            print("x[::-3] = ", x[::-3])
            print("x[:0:-1] = ", x[:0:-1])
            print("x[-5::-2] = ", x[-5::-2])
            print("x[:-5:-2] = ", x[:-5:-2])
            print("x[-2:-8:-2] = ", x[-2:-8:-2])
            print("x[0:9] = ", x[0:9])
            print("x[0:10:1] = ", x[0:10:1] )
            print("x[10:0] = ", x[10:0])
            print("x[10:0:-1] = ", x[10:0:-1])
            print("x[10:-10:-1] = ", x[10:-10:-1])
            print("x[10:-11:-1] = ", x[10:-11:-1])
            """)

    def test_count(self):
        self.assertCodeExecution("""
            print(b'abcabca'.count(97))
            print(b'abcabca'.count(b'abc'))
            print(b'qqq'.count(b'q'))
            print(b'qqq'.count(b'qq'))
            print(b'qqq'.count(b'qqq'))
            print(b'qqq'.count(b'qqqq'))
            print(b'abcdefgh'.count(b'bc',-7, -5))
            print(b'abcdefgh'.count(b'bc',1, -5))
            print(b'abcdefgh'.count(b'bc',0, 3))
            print(b'abcdefgh'.count(b'bc',-7, 500))
            print(b'qqaqqbqqqcqqqdqqqqeqqqqf'.count(b'qq'),1)
            print(b''.count(b'q'),0)
        """)
        self.assertCodeExecution("""
            b'abcabc'.count([]) #Test TypeError invalid byte array
        """, exits_early=True)
        self.assertCodeExecution("""
            b'abcabc'.count(256) #Test ValueError invalid integer range
        """, exits_early=True)
        self.assertCodeExecution("""
            print(b'abcabca'.count(97, [], 3)) #Test Slicing Error on Start
        """, exits_early=True)
        self.assertCodeExecution("""
            print(b'abcabca'.count(97, 3, [])) #Test Slicing Error on End
        """, exits_early=True)

    def test_find(self):
        self.assertCodeExecution("""
            print(b''.find(b'a'))
            print(b'abcd'.find(b''))
            print(b'abcd'.find(b'...'))
            print(b'abcd'.find(b'a'))
            print(b'abcd'.find(b'b'))
            print(b'abcd'.find(b'c'))
            print(b'abcd'.find(b'd'))
            print(b'abcd'.find(b'ab'))
            print(b'abcd'.find(b'bc'))
            print(b'abcd'.find(b'cd'))
            print(b'abcd'.find(b'cd', 2))
            print(b'abcd'.find(b'ab', 3))
            print(b'abcd'.find(b'cd', 2, 3))
            print(b'abcd'.find(b'ab', 3, 4))
        """)

    def test_index(self):
        self.assertCodeExecution("""
            print(b'abcd'.index(b'ab'))
            print(b'abcd'.index(b'bc'))
            print(b'abcd'.index(b'cd'))
            print(b'abcd'.find(b'cd', 2))
            print(b'abcd'.find(b'cd', 2, 3))
        """)
        self.assertCodeExecution("""
            print(b''.index(b'a'))
            print(b'abcd'.index(b''))
            print(b'abcd'.index(b'...'))
            print(b'abcd'.find(b'ab', 3))
            print(b'abcd'.find(b'ab', 3, 4))
        """, exits_early=True)

    def test_contains(self):
        self.assertCodeExecution("""
            print(b'py' in b'pybee')
            print(b'bee' in b'pybee')
            print(b'ybe' in b'pybee')
            print(b'test' in b'pybee')
            print(101 in b'pybee')
        """)
        self.assertCodeExecution("""
            print(300 in b'pybee') #Test ValueError invalid integer range
        """, exits_early=True)
        self.assertCodeExecution("""
            print(['b', 'e'] in b'pybee') #Test TypeError invalid byte array
        """, exits_early=True)

    def test_isalnum(self):
        self.assertCodeExecution("""
            print(b'w1thnumb3r2'.isalnum())
            print(b'withoutnumber'.isalnum())
            print(b'with spaces'.isalnum())
            print(b'666'.isalnum())
            print(b'66.6'.isalnum())
            print(b' '.isalnum())
            print(b''.isalnum())
            print(b'/@. test'.isalnum())
            print(b'\x46\x55\x43\x4B'.isalnum())
        """)

    def test_isalpha(self):
        self.assertCodeExecution("""
            print(b'testalpha'.isalpha())
            print(b'TestAlpha'.isalpha())
            print(b'test alpha'.isalpha())
            print(b'666'.isalpha())
            print(b'66.6'.isalpha())
            print(b' '.isalpha())
            print(b''.isalpha())
            print(b'/@. test'.isalpha())
            print(b'\x46\x55\x43\x4B'.isalpha())
        """)

    def test_isdigit(self):
        self.assertCodeExecution("""
            print(b'testdigit'.isdigit())
            print(b'TestDigit'.isdigit())
            print(b'test digit'.isdigit())
            print(b'666'.isdigit())
            print(b'66.6'.isdigit())
            print(b' '.isdigit())
            print(b''.isdigit())
            print(b'/@. test'.isdigit())
            print(b'\x46\x55\x43\x4B'.isdigit())
        """)

    def test_center(self):
        self.assertCodeExecution("""
            print(b'pybee'.center(12))
            print(b'pybee'.center(13))
            print(b'pybee'.center(2))
            print(b'pybee'.center(2, b'a'))
            print(b'pybee'.center(12, b'a'))
            print(b'pybee'.center(13, b'a'))
            print(b'pybee'.center(-5))
            print(b''.center(5))
            print(b'pybee'.center(True, b'a'))
            print(b'pybee'.center(True, bytearray(b'a')))
        """)
        self.assertCodeExecution("""
            print(b'pybee'.center('5'))
        """, exits_early=True)
        self.assertCodeExecution("""
            print(b'pybee'.center(12, b'as'))
        """, exits_early=True)
        self.assertCodeExecution("""
            print(b'pybee'.center(12, 'a'))
        """, exits_early=True)

    def test_upper(self):
        self.assertCodeExecution("""
            print(b'testupper'.upper())
            print(b'TestUpper'.upper())
            print(b'test upper'.upper())
            print(b'666'.upper())
            print(b' '.upper())
            print(b''.upper())
            print(b'/@. test'.upper())
            print(b'\x46\x55\x43\x4B'.upper())
        """)

    def test_ljust(self):
        self.assertCodeExecution("""
            print(b'testMoreThanWidth'.ljust(5))
            print(b'testEqualWidth'.ljust(14))
            print(b'testLessThanWidth'.ljust(20))
            print(b'testMoreWithFill'.ljust(2, b'x'))
            print(b'testEqualWithFill'.ljust(17, b'x'))
            print(b'testLessWithFill'.ljust(25, b'x'))
            print(b'testNegative'.ljust(-20))
            print(b''.ljust(5))
            print(b'testNoChangeWidthOne'.ljust(True, b'x'))
            print(b'testBArraySecondArg'.ljust(True, bytearray(b'x')))
            try:
                print(b'testStrArgError'.ljust('5'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testMoreLengthError'.ljust(12, b'as'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testStrFillingChar'.ljust(12, 'a'))
            except Exception as e:
                print(str(e))
        """)

    def test_rjust(self):
        self.assertCodeExecution("""
            print(b'testMoreThanWidth'.rjust(5))
            print(b'testEqualWidth'.rjust(14))
            print(b'testLessThanWidth'.rjust(20))
            print(b'testMoreWithFill'.rjust(2, b'x'))
            print(b'testEqualWithFill'.rjust(17, b'x'))
            print(b'testLessWithFill'.rjust(25, b'x'))
            print(b'testNegative'.rjust(-20))
            print(b''.rjust(5))
            print(b'testNoChangeWidthOne'.rjust(True, b'x'))
            print(b'testBArraySecondArg'.rjust(True, bytearray(b'x')))
            try:
                print(b'testStrArgError'.rjust('5'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testMoreLengthError'.rjust(12, b'as'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testStrFillingChar'.rjust(12, 'a'))
            except Exception as e:
                print(str(e))
        """)

    def test_expandtabs(self):
        self.assertCodeExecution("""
            print(b'testNoTabs'.expandtabs())
            print(b'test\t'.expandtabs())
            print(b'testDoubleTab\t\t'.expandtabs())
            print(b'testTab\tandText'.expandtabs())
            print(b'testTab\t'.expandtabs(4))
            print(b'testTab\t\t'.expandtabs(4))
            print(b'test\t\t'.expandtabs(-4))
            print(b''.expandtabs(5))
            try:
                print(b'testErrorChar\t'.expandtabs('a'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testErrorChars\t'.expandtabs('as'))
            except Exception as e:
                print(str(e))
            try:
                print(b'testErrorCharNum\t'.expandtabs('1'))
            except Exception as e:
                print(str(e))
        """)

    def test_lower(self):
        self.assertCodeExecution("""
            print(b"abc".lower())
            print(b"HELLO WORLD!".lower())
            print(b"hElLO wOrLd".lower())
            print(b"[Hello] World".lower())
            """)

    def test_swapcase(self):
        self.assertCodeExecution("""
            print(b"abc".swapcase())
            print(b"ABC".swapcase())
            print(b"HELLO WORLD!".swapcase())
            print(b"hElLO wOrLd".swapcase())
            print(b"[Hello] World".swapcase())
            """)

    def test_isspace(self):
        self.assertCodeExecution("""
            print(b'testisspace'.isspace())
            print(b'test isspace'.isspace())
            print(b' '.isspace())
            print(b''.isspace())
            print(b' \x46'.isspace())
            print(b'   \t\t'.isspace())
            print(b' \x0b'.isspace())
            print(b' \f'.isspace())
            print(b' \\n'.isspace())
            print(b' \\r'.isspace())
        """)

    def test_endswith(self):
        self.assertCodeExecution("""
            print(b"abc".endswith(b"c"))
            print(b"abc".endswith(b"bc"))
            print(b"abc".endswith(b"d"))
            print(b"".endswith(b"a"))
            print(b"abc".endswith(b""))
            print(b"abcde".endswith(b"de"))
            print(b"abcde".endswith(b"de", 2))
            print(b"abcde".endswith(b"de", 4))
            print(b"abcde".endswith(b"bc", 0, 3))
            print(b"abcde".endswith(b"abc", 0, 3))
            print(b"abcde".endswith(b"abc", 0, 4))
            print(b"abcde".endswith(b"abc", 1, 3))
            print(b"abcde".endswith(b"abc", 1, 4))
            print(b"abcde".endswith(b"abc", -1, 4))
            print(b"abcde".endswith(b"abc", 1, -1))
            print(b"abcde".endswith(b"abc", -6, -2))
            print(b"abcde".endswith((b"abc",)))
            print(b"abcde".endswith((b"abc", b"de")))
            print(b"abcde".endswith((b"abc", b"de", b"c")))
            print(b"abcde".endswith((b"de", b"d"), -4))
            print(b"abcde".endswith((b"de", b"d"), -4, 0))
            print(b"abcde".endswith((b"de", b"d"), -4, -1))
            print(b"abcde".endswith((b"da", b"aaa"), -4))
        """)

    def test_startswith(self):
        self.assertCodeExecution("""
            print(b"abc".startswith(b"a"))
            print(b"abc".startswith(b"bc"))
            print(b"abc".startswith(b"ab"))
            print(b"abc".startswith(b"d"))
            print(b"".startswith(b"a"))
            print(b"abc".startswith(b""))
            print(b"abcde".startswith(b"ab"))
            print(b"abcde".startswith(b"de", 3))
            print(b"abcde".startswith(b"abc", 3))
            print(b"abcde".startswith(b"de", 2))
            print(b"abcde".startswith(b"bc", 1, 7))
            print(b"abcde".startswith(b"abc", 0, 3))
            print(b"abcde".startswith(b"abc", 0, 4))
            print(b"abcde".startswith(b"abc", -10, 4))
            print(b"abcde".startswith(b"abc", -5, 4))
            print(b"abcde".startswith(b"abc", -4, 4))
            print(b"abcde".startswith(b"abc", 1, -1))
            print(b"abcde".startswith(b"abc", -6, -2))
            print(b"abcde".startswith((b"abc",)))
            print(b"abcde".startswith((b"abc", b"de")))
            print(b"abcde".startswith((b"de",)))
            print(b"abcde".startswith((b"de", b"c")))
            print(b"abcde".startswith((b"de", b"c", b"b", b"ab")))
            print(b"abcde".startswith((b"de", b"d"), -2))
            print(b"abcde".startswith((b"de", b"d"), -2, 0))
            print(b"abcde".startswith((b"de"), -2))
            print(b"abcde".startswith((b"de", b"d"), -2, -1))
            print(b"abcde".startswith((b"da", b"aaa"), -4))
        """)

    def test_strip(self):
        self.assertCodeExecution(r"""
            print(b"abcde".lstrip())
            print(b"      abcde".lstrip())
            print(b"\n   \t  abcde".lstrip())
            print(b"abcde".rstrip())
            print(b"abcde      ".rstrip())
            print(b"abcde   \t  \n".rstrip())
            print(b"abcde".strip())
            print(b"abcde      ".strip())
            print(b"    abcde   \t  \n".strip())
            print(b"".strip())
            print(b" \n".strip())
        """)

    def test_title(self):
        self.assertCodeExecution(r"""
            print(b"".title())
            print(b"abcd".title())
            print(b"NOT".title())
            print(b"coca cola".title())
            print(b"they are from UK, are they not?".title())
            print(b'/@.'.title())
            print(b'\x46\x55\x43\x4B'.title())
            print(b"py.bee".title())
        """)

    def test_istitle(self):
        self.assertCodeExecution(r"""
            print(b"".istitle())
            print(b"abcd".istitle())
            print(b"NOT".istitle())
            print(b"coca cola".istitle())
            print(b"they are from UK, are they not?".istitle())
            print(b'/@.'.istitle())
            print(b'\x46\x55\x43\x4B'.istitle())
            print(b"py.bee".title())
        """)

    def test_split(self):
        self.assertCodeExecution("""
            print(b''.split())
            print(b'py bee'.split())
            print(b'pyXbXee'.split(b'X'))
            print(b'pyXbee'.split(b'z'))
            print(b'pyZZbee'.split(b'ZZ'))
            print(b'pybebyp'.split(b'e', 1))
            print(b'aabaabaa'.split(b'b', 1))
            print(b'aabaabaa'.split(b'b', -1))
            print(b'one two three'.split(maxsplit=1))
            """)
        self.assertCodeExecution("""
            print(b''.split('a'))
            """, exits_early=True)
        self.assertCodeExecution("""
            print(b'pyXbee'.split('a'))
            """, exits_early=True)
        self.assertCodeExecution("""
            print(b'pyXbee'.split(maxsplit='5'))
            """, exits_early=True)
        self.assertCodeExecution("""
            print(b''.split(maxsplit='5'))
            """, exits_early=True)

    def test_join(self):
        self.assertCodeExecution("""
            b = bytes(b'.')
            print(b.join([b'12', b'dh']))
            print(b.join([bytearray(b'12'), bytearray(b'dh')]))
            b = bytes(b' ')
            print(b.join([b'd', bytearray(b'l22-'), b'=ej*']))
            print(b.join([bytearray(b'31'), b'`', b'^']))
            print(b.join([bytearray(b'dh')]))
            b = bytes(b'%#@!')
            print(b.join([b'1',b'd',b'<']))
            print(b.join([b'12']))
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(bytes(1, 2, 3, 4))
            except TypeError as err:
                print(err)
        """)


class UnaryBytesOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'bytes'

    not_implemented = [
    ]


class BinaryBytesOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'bytes'

    not_implemented = [
        'test_direct_eq_bytearray',
        'test_direct_eq_none',

        'test_direct_ge_bytearray',
        'test_direct_ge_none',

        'test_direct_gt_bytearray',
        'test_direct_gt_none',

        'test_direct_le_bytearray',
        'test_direct_le_none',

        'test_direct_lt_bytearray',
        'test_direct_lt_none',

        'test_direct_ne_bytearray',
        'test_direct_ne_none',

        'test_modulo_complex',
        'test_modulo_dict',
    ]

    not_implemented_versions = {
        'test_modulo_None': (3.5, 3.6),
        'test_modulo_NotImplemented': (3.5, 3.6),
        'test_modulo_bool': (3.5, 3.6),
        'test_modulo_bytearray': (3.5, 3.6),
        'test_modulo_bytes': (3.5, 3.6),
        'test_modulo_class': (3.5, 3.6),
        'test_modulo_float': (3.5, 3.6),
        'test_modulo_frozenset': (3.5, 3.6),
        'test_modulo_int': (3.5, 3.6),
        'test_modulo_list': (3.5, 3.6),
        'test_modulo_range': (3.5, 3.6),
        'test_modulo_set': (3.5, 3.6),
        'test_modulo_slice': (3.5, 3.6),
        'test_modulo_str': (3.5, 3.6),
        'test_modulo_tuple': (3.5, 3.6),
        'test_modulo_obj': (3.5, 3.6),
    }


class InplaceBytesOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'bytes'

    not_implemented = [
        'test_modulo_complex',
    ]

    not_implemented_versions = {
        'test_modulo_None': (3.5, 3.6),
        'test_modulo_NotImplemented': (3.5, 3.6),
        'test_modulo_bool': (3.5, 3.6),
        'test_modulo_bytearray': (3.5, 3.6),
        'test_modulo_bytes': (3.5, 3.6),
        'test_modulo_class': (3.5, 3.6),
        'test_modulo_float': (3.5, 3.6),
        'test_modulo_frozenset': (3.5, 3.6),
        'test_modulo_int': (3.5, 3.6),
        'test_modulo_list': (3.5, 3.6),
        'test_modulo_range': (3.5, 3.6),
        'test_modulo_set': (3.5, 3.6),
        'test_modulo_slice': (3.5, 3.6),
        'test_modulo_str': (3.5, 3.6),
        'test_modulo_tuple': (3.5, 3.6),
        'test_modulo_obj': (3.5, 3.6),
    }

    is_flakey = [
        'test_modulo_dict',
    ]



from unittest import expectedFailure

from ..utils import TranspileTestCase


class NonlocalTests(TranspileTestCase):
    def test_nonlocal_func(self):
        self.assertCodeExecution("""
            def func():
                a = 'a from outer'
                b = 'b from outer'
                def nested_func():
                    nonlocal a
                    print(a)
                    a = 'a from inner'
                    print(a)
                    b = 'b from inner'
                    print(b)

                nested_func()
                print(a)
                print(b)

            func()

            def func2():
                a = 'a from outer'
                b = 'b from outer'
                def nested_func2():
                    nonlocal a
                    print(a)
                    a = 'a from inner'
                    print(a)
                    def nested_nested_func():
                        nonlocal b
                        print(b)
                        b = 'b from innest'
                        print(b)

                    nested_nested_func()

                nested_func2()
                print(a)
                print(b)

            func2()
        """)

        self.assertCodeExecution("""
            def func():
                a = None
                def nested():
                    nonlocal a
                    a = 'changed by nested'
                    print(a)

                def nested2():
                    print(a)

                return (nested, nested2)

            nested, nested2 = func()
            nested2()
            nested()
            nested2()
        """)

    @expectedFailure
    def test_nonlocal_class(self):
        self.assertCodeExecution("""
            def func():
                a = 'a from outer'
                b = 'b from outer'
                class Inner():
                    nonlocal a
                    print(a)
                    a = 'a from inner'
                    b = 'b from inner'
                    print(a)
                    print(b)

                Inner()
                print(a)
                print(b)

            func()
        """)

    @expectedFailure
    def test_nonlocal_method(self):
        self.assertCodeExecution("""
            def func():
                a = 'a from outer'
                b = 'b from outer'
                class Klass:
                    def method(self):
                        nonlocal a
                        print(a)
                        a = 'a from inner'
                        print(a)
                        print(b)

                Klass().method()
                print(a)
                print(b)

            func()
        """)

    def test_nonlocal_generator(self):
        self.assertCodeExecution("""
            def func():
                a = 'a from outer'
                b = 'b from outer'
                def gen():
                    nonlocal a
                    print(a)
                    print(b)
                    a = 'a from inner'
                    yield a

                print(next(gen()))
                print(a)
                print(b)

            func()

            def func2():
                a = 'a from outer'
                b = 'b from outer'
                def gen():
                    nonlocal a
                    print(a)
                    print(b)
                    a = 'a from inner'
                    yield a

                print(next(gen()))
                print(a)
                print(b)
                yield

            next(func2())
        """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class DictTests(TranspileTestCase):
    pass


class BuiltinDictFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["dict"]



import sys
from ..java import (
    Annotation, Code as JavaCode, ConstantElementValue, Method as JavaMethod,
    RuntimeVisibleAnnotations, opcodes as JavaOpcodes, Classref as JavaClassref
)
from .blocks import Block, Accumulator, BlockCodeTooLarge
from .structures import (
    TRY, CATCH, END_TRY,
    ArgType, IF, END_IF
)
from .types import java, python
from .types.primitives import (
    ALOAD_name, ASTORE_name, free_name,
    DLOAD_name, FLOAD_name,
    ICONST_val, ILOAD_name,
    LLOAD_name,
)
# from .debug import DEBUG, DEBUG_value


CO_VARARGS = 0x0004
CO_VARKEYWORDS = 0x0008
CO_GENERATOR = 0x0020


def descriptor(annotation):
    if annotation == 'bool':
        return 'Z'
    elif annotation == 'byte':
        return 'B'
    elif annotation == 'char':
        return 'C'
    elif annotation == 'short':
        return 'S'
    elif annotation == 'int':
        return 'I'
    elif annotation == 'long':
        return 'J'
    elif annotation == 'float':
        return 'F'
    elif annotation == 'double':
        return 'D'
    elif annotation is None or annotation == 'void':
        return 'V'
    else:
        return 'L%s;' % annotation.replace('.', '/')


def to_python(accumulator, annotation, var_name):
    if annotation == "bool":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            ILOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Bool',
                'getBool',
                args=['Ljava/lang/Long;'],
                returns='Lorg/python/types/Bool;'),
        )
    elif annotation == "byte":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            ILOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Int',
                'getInt',
                args=['B'],
                returns='Lorg/python/types/Int;'
            ),
        )
    elif annotation == 'char':
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            ILOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Int',
                'getInt',
                args=['C'],
                returns='Lorg/python/types/Int;'
            ),
        )
    elif annotation == "short":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            ILOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Int',
                'getInt',
                args=['S'],
                returns='Lorg/python/types/Int;'
            ),
        )
    elif annotation == "int":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            ILOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Int',
                'getInt',
                args=['I'],
                returns='Lorg/python/types/Int;'
            ),
        )
    elif annotation == "long":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            LLOAD_name(var_name),
            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Int',
                'getInt',
                args=['J'],
                returns='Lorg/python/types/Int;'
            ),
        )
    elif annotation == "float":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            java.New('org/python/types/Float'),
            FLOAD_name(var_name),
            java.Init('org/python/types/Float', 'F'),
        )
    elif annotation == "double":
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),

            java.New('org/python/types/Float'),
            DLOAD_name(var_name),
            java.Init('org/python/types/Float', 'D'),
        )
    else:
        accumulator.add_opcodes(
            # DEBUG("INPUT %s TRANSFORM %s" % (i, annotation)),
            ALOAD_name(var_name),
            python.Type.to_python(),
        )


def to_java(accumulator, annotation):
    if annotation == 'void':
        accumulator.add_opcodes(
            JavaOpcodes.POP(),
        )
    elif annotation == 'bool':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Bool'),
            JavaOpcodes.GETFIELD('org/python/types/Bool', 'value', 'Z'),
        )
    elif annotation == 'byte':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Int'),
            JavaOpcodes.GETFIELD('org/python/types/Int', 'value', 'J'),
            JavaOpcodes.L2I(),
            JavaOpcodes.I2B(),
        )
    elif annotation == 'char':
        accumulator.add_opcodes(
            JavaOpcodes.INVOKEINTERFACE('org/python/Object', 'toJava', args=[], returns='Ljava/lang/Object;'),
            JavaOpcodes.CHECKCAST('java/lang/String'),
            ICONST_val(0),
            JavaOpcodes.INVOKEVIRTUAL('java/lang/String', 'charAt', args=['I'], returns='C'),
        )
    elif annotation == 'short':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Int'),
            JavaOpcodes.GETFIELD('org/python/types/Int', 'value', 'J'),
            JavaOpcodes.L2I(),
            JavaOpcodes.I2S(),
        )
    elif annotation == 'int':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Int'),
            JavaOpcodes.GETFIELD('org/python/types/Int', 'value', 'J'),
            JavaOpcodes.L2I(),
        )
    elif annotation == 'long':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Int'),
            JavaOpcodes.GETFIELD('org/python/types/Int', 'value', 'J'),
        )
    elif annotation == 'float':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Float'),
            JavaOpcodes.GETFIELD('org/python/types/Float', 'value', 'D'),
            JavaOpcodes.DTOF(),
        )
    elif annotation == 'double':
        accumulator.add_opcodes(
            JavaOpcodes.CHECKCAST('org/python/types/Float'),
            JavaOpcodes.GETFIELD('org/python/types/Float', 'value', 'D'),
        )
    elif annotation != 'org/python/Object':
        accumulator.add_opcodes(
            JavaOpcodes.INVOKEINTERFACE('org/python/Object', 'toJava', args=[], returns='Ljava/lang/Object;'),
            JavaOpcodes.CHECKCAST(annotation.replace('.', '/')),
        )


def return_statement(accumulator, annotation):
    if annotation == 'void':
        accumulator.add_opcodes(
            JavaOpcodes.RETURN()
        )
    elif annotation in ['bool', 'byte', 'char', 'short', 'int']:
        accumulator.add_opcodes(
            JavaOpcodes.IRETURN(),
        )
    elif annotation == 'long':
        accumulator.add_opcodes(
            JavaOpcodes.LRETURN(),
        )
    elif annotation == 'float':
        accumulator.add_opcodes(
            JavaOpcodes.FRETURN(),
        )
    elif annotation == 'double':
        accumulator.add_opcodes(
            JavaOpcodes.DRETURN(),
        )
    else:
        accumulator.add_opcodes(
            JavaOpcodes.ARETURN()
        )


class MethodCodeTooLarge(Exception):
    pass


class Function(Block):
    def __init__(self, parent, name, code, parameters, returns, static=False):
        super().__init__(parent=parent)
        self.name = name

        # Python can redefine function symbols. Keep a track of any
        # function that is defined; if the symbol has already been
        # defined in this context, then append $n to the symbol in
        # the Java classfile.
        duplicates = self._parent.symbols.setdefault(self.name, [])
        duplicates.append(self.method_name)
        if len(duplicates) > 1:
            self._java_suffix = '$%d' % (len(duplicates) - 1)
        else:
            self._java_suffix = ''

        self.code = code
        self.parameters = parameters
        self.returns = returns

        # Reserve space for the register that will hold self (if required)
        self.add_self()

        # Reserve space for the registers that will hold arguments.
        for i, param in enumerate(self.parameters):
            self.local_vars[param['name']] = len(self.local_vars)

        # Reserve space for the register that will hold locals
        self.local_vars['#locals'] = len(self.local_vars)

        self.static = static

    def __repr__(self):
        return '<Function %s (%s parameters)>' % (self.name, len(self.parameters))

    @property
    def is_constructor(self):
        return False

    def add_self(self):
        pass

    def store_module(self):
        # Stores the current module as a local variable
        if ('#module') not in self.local_vars:
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),

                ASTORE_name('#module'),
            )

    def store_name(self, name, declare=False):
        if declare or name in self.local_vars:
            self.add_opcodes(
                # Store in a local variable
                ASTORE_name(name),
            )
            self.add_opcodes(
                # Also store in the locals variable
                ALOAD_name('#locals'),
                JavaOpcodes.LDC_W(name),
                ALOAD_name(name),
                java.Map.put(),
            )
        else:
            self.add_opcodes(
                ASTORE_name('#value'),

                ALOAD_name('#module'),

                ALOAD_name('#value'),

                python.Object.set_attr(name),
                free_name('#value')
            )

    def store_dynamic(self):
        raise NotImplementedError('Functions cannot dynamically store variables.')

    def load_name(self, name):
        if name in self.local_vars:
            self.add_opcodes(
                ALOAD_name('#locals'),
                java.Map.get(name),
                JavaOpcodes.DUP(),
                IF([], JavaOpcodes.IFNONNULL),
                java.THROW(
                    'org/python/exceptions/UnboundLocalError',
                    ['Ljava/lang/String;', JavaOpcodes.LDC_W(name)]
                ),
                END_IF(),
            )
        else:
            self.add_opcodes(
                ALOAD_name('#module'),
                python.Object.get_attribute(name),
            )

    def load_globals(self):
        self.add_opcodes(
            ALOAD_name('#module'),
            JavaOpcodes.GETFIELD('org/python/types/Module', '__dict__', 'Ljava/util/Map;'),
        )

    def load_locals(self):
        self.add_opcodes(
            ALOAD_name('#locals')
        )

    def load_vars(self):
        self.load_locals()

    def delete_name(self, name):
        try:
            self.add_opcodes(
                free_name(name),
                ALOAD_name('#locals'),
                java.Map.remove(name)
            )
        except NameError:
            self.add_opcodes(
                ALOAD_name('#module'),
                python.Object.del_attr(name),
            )

    @property
    def can_ignore_empty(self):
        return False

    @property
    def signature(self):
        return_descriptor = 'Lorg/python/Object;'
        return '(%s)%s' % (
            ''.join('Lorg/python/Object;' for p in self.parameters),
            return_descriptor
        )

    @property
    def method_name(self):
        return self.name

    @property
    def java_name(self):
        return self.name + self._java_suffix

    @property
    def pyimpl_name(self):
        return self.name + self._java_suffix

    @property
    def module(self):
        return self._parent

    @property
    def class_descriptor(self):
        return self.module.class_descriptor

    def add_class(self, class_name, extends, implements):
        from .klass import Class

        klass = Class(
            self.module,
            name=class_name,
            extends=extends,
            implements=implements
        )

        self.module.classes.append(klass)

        self.add_opcodes(
            # Stack contains the bases list
            ASTORE_name('#bases'),

            # DEBUG("FORCE LOAD OF CLASS %s AT DEFINITION" % klass.descriptor),
            # - class
            JavaOpcodes.LDC_W(JavaClassref(klass.descriptor)),

            # - name
            JavaOpcodes.LDC_W(klass.name),

            # - bases
            ALOAD_name('#bases'),

            # - dict
            JavaOpcodes.ACONST_NULL(),

            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Type',
                'declarePythonType',
                args=[
                    'Ljava/lang/Class;',
                    'Ljava/lang/String;',
                    'Ljava/util/List;',
                    'Ljava/util/Map;'
                ],
                returns='Lorg/python/types/Type;'
            ),

            free_name('#bases')
        )

        self.store_name(klass.name)

        self.add_opcodes(
            JavaOpcodes.INVOKESTATIC(
                klass.descriptor,
                'class$init',
                args=[],
                returns='V'
            ),
        )

        return klass

    def add_function(self, name, code, parameter_signatures, return_signature):
        # If a function is added to a function, it is added as an anonymous
        # inner class to the function/method's parent module/class.
        from .klass import ClosureClass
        if not (isinstance(self, Closure) or isinstance(self, GeneratorClosure)):
            # prepend name to first level nested function
            name = self.name + '$' + name

        klass = ClosureClass(
            parent=self._parent,
            name=name,
        )
        self.module.classes.append(klass)

        klass.visitor_setup()

        if hasattr(self, "outer_contexts") and self.outer_contexts:
            outer_contexts = self.outer_contexts + [self]
        else:
            outer_contexts = [self]

        if code.co_flags & CO_GENERATOR:
            closure = GeneratorClosure(
                klass,
                code=code,
                generator=code.co_name,
                parameters=parameter_signatures,
                returns=return_signature,
                outer_contexts=outer_contexts
            )
        else:
            closure = Closure(
                klass,
                code=code,
                parameters=parameter_signatures,
                returns=return_signature,
                outer_contexts=outer_contexts
            )

        klass.methods.append(closure)

        for field in code.co_names:
            klass.fields[field] = 'Lorg/python/Object;'

        klass.visitor_teardown()

        # closure has reference to outer context's local variables by maintaining a list of #locals
        self.add_opcodes(
            java.New(klass.descriptor),
            java.List(),
            JavaOpcodes.DUP(),
            ALOAD_name('#locals'),
            java.List.add(),
        )

        if isinstance(self, Closure):
            self.add_opcodes(
                JavaOpcodes.DUP(),
                ALOAD_name('<closure>'),
                JavaOpcodes.CHECKCAST('org/python/types/Closure'),
                JavaOpcodes.GETFIELD('org/python/types/Closure', 'locals_list', 'Ljava/util/List;'),
                java.List.addAll(),
            )
        elif isinstance(self, GeneratorClosure):
            self.add_opcodes(
                JavaOpcodes.DUP(),
                ALOAD_name('<generator>'),
                JavaOpcodes.CHECKCAST('org/python/types/Generator'),
                JavaOpcodes.GETFIELD('org/python/types/Generator', 'closure', 'Lorg/python/types/Closure;'),
                JavaOpcodes.GETFIELD('org/python/types/Closure', 'locals_list', 'Ljava/util/List;'),
                java.List.addAll(),
            )

        self.add_opcodes(
            java.Init(klass.descriptor, 'Ljava/util/List;'),

            python.Type.for_name(klass.descriptor),
        )

        self.add_callable(closure)

        self.add_opcodes(
            python.Object.set_attr('invoke'),
            python.Object.get_attribute('invoke'),
        )

        return closure

    def visitor_setup(self):
        self.add_opcodes(
            java.Map(),
            ASTORE_name('#locals')
        )

        # stores parameters in #locals
        for param in self.parameters:
            self.add_opcodes(
                ALOAD_name('#locals'),
                JavaOpcodes.LDC_W(param['name']),
                ALOAD_name(param['name']),
                java.Map.put(),
            )

        self.store_module()

    def visitor_teardown(self):
        if len(self.opcodes) == 0:
            # If there is no content in this method, add a RETURN.
            return_required = True
        elif isinstance(self.opcodes[-1], (JavaOpcodes.RETURN, JavaOpcodes.ARETURN)):
            # If the last opcode in the method is a RETURN, but that
            # return was not added at the root level (i.e., it's not a return
            # as the last statement of an IF statement), add one.
            try:
                return_required = self.opcodes[-1].needs_implicit_return
            except AttributeError:
                return_required = True
        else:
            # If the last statement in the block isn't a RETURN, add one.
            return_required = True

        if return_required:
            self.add_opcodes(
                python.NONE(),
                JavaOpcodes.ARETURN()
            )

    def method_attributes(self):
        return [
            RuntimeVisibleAnnotations([
                Annotation(
                    'Lorg/python/Method;',
                    {
                        '__doc__': ConstantElementValue("Python method (insert docs here)")
                    }
                )
            ])
        ]

    def transpile_method(self):
        try:
            return [
                JavaMethod(
                    self.pyimpl_name,
                    self.signature,
                    static=self.static,
                    attributes=[self.transpile_code()] + self.method_attributes()
                )
            ]
        except BlockCodeTooLarge as e:
            raise MethodCodeTooLarge("Code is too large for method %s: %d > 65534"
                                     % (self.name, e.code_length))

    def transpile_wrapper(self):
        return []

    def transpile(self):
        return self.transpile_method() + self.transpile_wrapper()


class InitMethod(Function):
    def __init__(self, klass, args=None, super_args=None, parameters=None):
        super().__init__(
            klass,
            name='<init>',
            code=None,
            parameters=parameters if parameters else [
                {
                    'name': 'self',
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'annotation': 'org/python/Object'
                },
                {
                    'name': 'args',
                    'kind': ArgType.VAR_POSITIONAL,
                    'annotation': '[Lorg/python/Object;'
                },
                {
                    'name': 'kwargs',
                    'kind': ArgType.VAR_KEYWORD,
                    'annotation': 'java/util/Map'
                }
            ],
            returns={'annotation': None},
        )
        self.args = args if args else {}
        self.super_args = super_args if super_args else []

        self.store_module()

    def __repr__(self):
        return '<Constructor %s (%s parameters)>' % (self.klass.name, len(self.parameters))

    @property
    def is_constructor(self):
        return True

    @property
    def klass(self):
        return self._parent

    @property
    def module(self):
        return self.klass.module

    @property
    def class_descriptor(self):
        return self.klass.descriptor

    @property
    def can_ignore_empty(self):
        return False

    @property
    def signature(self):
        return '([Lorg/python/Object;Ljava/util/Map;)V'

    def load_name(self, name):
        if name in self.args:
            index = self.args[name]
            if index is None:
                self.add_opcodes(
                    JavaOpcodes.ALOAD_2(),
                    java.Map.get(name)
                )
            else:
                self.add_opcodes(
                    JavaOpcodes.ALOAD_1(),
                    java.Array.get(index),
                )
        else:
            self.add_opcodes(
                ALOAD_name('#module'),
                python.Object.get_attribute(name),
            )

    def visitor_setup(self):
        # Construct the contents of the constructor method...
        self.add_opcodes(
            # Create the instance
            JavaOpcodes.ALOAD_0(),
            JavaOpcodes.DUP(),
        )

    def visitor_teardown(self):
        self.add_opcodes(
            java.Init(self.klass.extends_descriptor, *[descriptor(arg) for arg in self.super_args]),
            python.Type.to_python(),

            python.Object.get_attribute('__init__', use_null=True),
            # If no __init__ exists, just return.
            JavaOpcodes.DUP(),
            JavaOpcodes.IFNULL(13),  # 3

            # Check that it is a callable
            JavaOpcodes.CHECKCAST('org/python/Callable'),  # 3

            # ...and invoke it
            JavaOpcodes.ALOAD_1(),  # 1
            JavaOpcodes.ALOAD_2(),  # 1
            python.Callable.invoke(),  # 5

            JavaOpcodes.POP(),  # 1

            JavaOpcodes.RETURN()
        )


class Method(Function):
    def __init__(self, klass, name, code, parameters, returns=None, static=False):
        super().__init__(
            klass,
            name=name,
            code=code,
            parameters=parameters,
            returns=returns,
            static=static,
        )

        self.store_module()

    def __repr__(self):
        return '<Method %s.%s (%s parameters)>' % (self.klass.name, self.name, len(self.parameters))

    @property
    def pyimpl_name(self):
        return self.java_name + "$py" + self._java_suffix

    @property
    def klass(self):
        return self._parent

    @property
    def module(self):
        return self.klass.module

    @property
    def class_descriptor(self):
        return self.klass.descriptor

    @property
    def bound_signature(self):
        return_descriptor = descriptor(self.returns.get('annotation'))
        return '(%s)%s' % (
            ''.join(descriptor(p['annotation']) for p in self.parameters[1:]),
            return_descriptor
        )

    def transpile_wrapper(self):
        # The first register holds "self"; since the binding is only
        # invoked via a native call, wrap it in a Java Object - unless
        # the object is an extension type, in which case the wrapper
        # object should already exist.
        binding = Accumulator(self.local_vars)

        binding.add_opcodes(
            # DEBUG("BINDING FOR " + self.name + self.signature),
            # DEBUG("BOUND AS " + self.name + self.bound_signature),
            # JavaOpcodes.ALOAD_0(),
            # DEBUG_value("BINDING SELF IN"),

            JavaOpcodes.ALOAD_0(),
            python.Type.to_python(),  # 3

            # DEBUG_value("BINDING SELF OUT", dup=True),
        )

        # Then extract each argument, converting to Python types as required.
        for i, param in enumerate(self.parameters[1:]):
            if param['annotation'] is None:
                raise Exception("Parameters can't be void")
            else:
                to_python(binding, param['annotation'], param['name'])
            # binding.add_opcodes(
            #     DEBUG("INPUT %s TRANSFORMED" % (i)),
            # )
        # binding.add_opcodes(
        #     DEBUG("INPUTS TRANSFORMED"),
        # )

        # Then call the method, and process the return type.
        binding.add_opcodes(
            JavaOpcodes.INVOKESTATIC(self.klass.descriptor, self.pyimpl_name, self.signature),
        )

        to_java(binding, self.returns['annotation'])
        return_statement(binding, self.returns['annotation'])

        # binding.add_opcodes(
        #     DEBUG("BINDING OUTPUT to type %s" % self.returns['annotation']),
        # )

        wrapper_methods = [
            JavaMethod(
                self.java_name,
                self.bound_signature,
                attributes=[
                    JavaCode(
                        max_stack=len(self.parameters) + 5,
                        max_locals=len(self.parameters),
                        code=binding.opcodes
                    )
                ]
            ),
        ]

        if self.klass.extends:
            # We also need to generate a manual call to the super
            super_wrapper = Accumulator()
            super_wrapper.add_opcodes(
                JavaOpcodes.ALOAD_0(),
            )

            # Then extract each argument:
            for i, param in enumerate(self.parameters[1:]):
                super_wrapper.add_opcodes({
                        'bool': JavaOpcodes.ILOAD(i + 1),
                        'byte': JavaOpcodes.ILOAD(i + 1),
                        'char': JavaOpcodes.ILOAD(i + 1),
                        'short': JavaOpcodes.ILOAD(i + 1),
                        'int': JavaOpcodes.ILOAD(i + 1),
                        'long': JavaOpcodes.ILOAD(i + 1),
                        'float': JavaOpcodes.FLOAD(i + 1),
                        'double': JavaOpcodes.DLOAD(i + 1),
                    }.get(param['annotation'], JavaOpcodes.ALOAD(i + 1))
                )

            # Then call the method, and process the return type.
            super_wrapper.add_opcodes(
                JavaOpcodes.INVOKESPECIAL(self.klass.extends.replace('.', '/'), self.name, self.bound_signature),

                {
                    'void': JavaOpcodes.RETURN(),
                    'bool': JavaOpcodes.IRETURN(),
                    'byte': JavaOpcodes.IRETURN(),
                    'char': JavaOpcodes.IRETURN(),
                    'short': JavaOpcodes.IRETURN(),
                    'int': JavaOpcodes.IRETURN(),
                    'long': JavaOpcodes.LRETURN(),
                    'float': JavaOpcodes.FRETURN(),
                    'double': JavaOpcodes.DRETURN(),
                }.get(self.returns['annotation'], JavaOpcodes.ARETURN()),

                # DEBUG("SUPER WRAPPER OUTPUT to type %s" % self.returns['annotation']),
            )

            wrapper_methods.append(
                JavaMethod(
                    self.java_name + "$super",
                    self.bound_signature,
                    attributes=[
                        JavaCode(
                            max_stack=len(self.parameters) + 2,
                            max_locals=len(self.parameters) + 2,
                            code=super_wrapper.opcodes
                        )
                    ]
                )
            )

        return wrapper_methods


class MainFunction(Function):
    def __init__(self, module):
        super().__init__(
            module,
            name='__main__',
            code=None,
            parameters=[{'name': 'args', 'annotation': 'argv'}],
            returns={'annotation': None},
            static=True,
        )

    def __repr__(self):
        return '<MainFunction %s>' % self.module.name

    @property
    def java_name(self):
        return 'main'

    @property
    def pyimpl_name(self):
        return self.java_name

    @property
    def module(self):
        return self._parent

    @property
    def signature(self):
        return '([Ljava/lang/String;)V'

    @property
    def can_ignore_empty(self):
        return True

    def store_name(self, name, declare=False):
        self.add_opcodes(
            ASTORE_name('#value'),
            ALOAD_name('#module'),  # #module is available as a local var after visitor_setup has been called

            ALOAD_name('#value'),
            python.Object.set_attr(name),
            free_name('#value')
        )

    def store_dynamic(self):
        self.add_opcodes(
            ASTORE_name('#value'),
            python.Type.for_name(self.module.class_name),
            JavaOpcodes.GETFIELD('org/python/types/Type', '__dict__', 'Ljava/util/Map;'),
            ALOAD_name('#value'),

            java.Map.putAll(),
            free_name('#value')
        )

    def load_name(self, name):
        self.add_opcodes(
            ALOAD_name('#module'),  # #module is available as a local var after visitor_setup has been called
            python.Object.get_attribute(name),
        )

    def delete_name(self, name):
        self.add_opcodes(
            ALOAD_name('#module'),  # #module is available as a local var after visitor_setup has been called
            python.Object.del_attr(name),
        )

    def visitor_setup(self):
        self.add_opcodes(
            # Add a TRY-CATCH for SystemExit
            TRY(),
        )
        self.add_opcodes(
                # Set the Python version
                JavaOpcodes.LDC_W(sys.hexversion),
                JavaOpcodes.PUTSTATIC('org/Python', 'VERSION', 'I'),

                # Construct this module
                java.New(self.module.class_descriptor),
                java.Init(self.module.class_descriptor),
                ASTORE_name('#module'),

                # Register the module by it's full name
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),
                python.Str(self.module.full_name),
                ALOAD_name('#module'),
                python.Object.set_item(),

                # Register the same module as __main__
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),
                python.Str('__main__'),
                ALOAD_name('#module'),
                python.Object.set_item(),

                # Set the module's __name__
                ALOAD_name('#module'),
                python.Str("__main__"),
                python.Object.set_attr('__name__'),

                # Set the module's __package__
                ALOAD_name('#module'),
                python.NONE(),
                python.Object.set_attr("__package__"),

                # Run the module block.
                ALOAD_name('#module'),
                JavaOpcodes.INVOKEVIRTUAL(self.module.class_descriptor, 'module$import', args=[], returns='V'),
        )

    def visitor_teardown(self):
        # Close out the TRY-CATCH for SystemExit
        self.add_opcodes(
            CATCH('org/python/exceptions/SystemExit'),
        )
        self.add_opcodes(
                JavaOpcodes.GETFIELD('org/python/exceptions/SystemExit', 'return_code', 'I'),
                JavaOpcodes.INVOKESTATIC('java/lang/System', 'exit', args=['I'], returns='V'),
        )
        self.add_opcodes(
            END_TRY(),
            JavaOpcodes.RETURN()
        )

    def method_attributes(self):
        return []


def _get_enclosing_context_level(child_context, name):
    # returns level of enclosing context that defined the variable `name`
    # i.e. level = 2 means `name` is found two levels up from the nested `child_context`
    if name in child_context.local_vars:
        return None
    else:
        level = 0
        for context in child_context.outer_contexts[::-1]:
            level += 1
            if name in context.local_vars and context.local_vars[name] is not None:
                return level

    return None


class Closure(Function):
    def __init__(self, klass, code, parameters, returns=None, static=False, outer_contexts=None):
        super().__init__(
            klass,
            name='invoke',
            code=code,
            parameters=parameters,
            returns=returns,
            static=static,
        )
        self.nonlocal_vars = []  # holds nonlocal variable names for `store_name`
        self.outer_contexts = outer_contexts  # parent scopes of this closure, excluding global scope

        self.store_module()

    def __repr__(self):
        return '<Closure %s (%s parameters)>' % (self.name, len(self.parameters))

    def store_name(self, name, declare=False):
        if name in self.nonlocal_vars:
            # updates closure
            self.add_opcodes(
                ALOAD_name('<closure>'),
                JavaOpcodes.CHECKCAST('org/python/types/Closure'),
                ICONST_val(_get_enclosing_context_level(self, name)),
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Closure',
                    'get_locals',
                    args=['I'],
                    returns='Ljava/util/Map;'
                ),
                JavaOpcodes.SWAP(),
                JavaOpcodes.LDC_W(name),
                JavaOpcodes.SWAP(),
                java.Map.put()
            )
        else:
            super().store_name(name, declare)

    def load_name(self, name):
        parent_level = _get_enclosing_context_level(self, name)
        if parent_level:
            self.add_opcodes(
                ALOAD_name('<closure>'),
                JavaOpcodes.CHECKCAST('org/python/types/Closure'),
                ICONST_val(parent_level),
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Closure',
                    'get_locals',
                    args=['I'],
                    returns='Ljava/util/Map;'
                ),
                java.Map.get(name),
            )
        else:
            super().load_name(name)

    @property
    def klass(self):
        return self._parent

    @property
    def module(self):
        return self.klass.module

    @property
    def class_descriptor(self):
        return self.klass.descriptor

    def add_self(self):
        # In a closure method, the first argument is the closure,
        # not self. self *might* be the first argument -- but in
        # that case, it's just the first argument in an unbound
        # method.
        self.local_vars['<closure>'] = len(self.local_vars)
        self.has_self = True


class ClosureInitMethod(InitMethod):
    def __init__(self, klass):
        super().__init__(
            klass,
            parameters=[
                {
                    'name': 'self',
                    'kind': ArgType.POSITIONAL_OR_KEYWORD,
                    'annotation': 'org/python/Object'
                },
                {
                    'name': 'kwargs',
                    'kind': ArgType.VAR_KEYWORD,
                    'annotation': 'java/util/Map'
                }
            ],
        )

    def __repr__(self):
        return '<Closure constructor %s (%s parameters)>' % (self.klass.name, len(self.parameters))

    @property
    def signature(self):
        return '(Ljava/util/List;)V'

    def visitor_teardown(self):
        self.add_opcodes(
            JavaOpcodes.ALOAD_1(),
            java.Init(self.klass.extends_descriptor, 'Ljava/util/List;'),

            JavaOpcodes.RETURN()
        )


class GeneratorFunction(Function):
    def __init__(self, module, name, code, generator, parameters, returns=None, static=False):
        super().__init__(
            module,
            name=name,
            code=code,
            parameters=parameters,
            returns=returns,
            static=static,
        )
        self.generator = generator

    @property
    def klass(self):
        return self.module

    def add_self(self):
        self.local_vars['<generator>'] = len(self.local_vars)
        self.has_self = True

    def visitor_setup(self):
        self.add_opcodes(
            # Restore the variables needed for the entry of the generator.
            ALOAD_name('<generator>'),
            JavaOpcodes.GETFIELD('org/python/types/Generator', 'stack', 'Ljava/util/Map;'),
            ASTORE_name('#locals'),
        )

        for param in self.parameters:
            self.add_opcodes(
                ALOAD_name('#locals'),
                java.Map.get(param['name']),
                ASTORE_name(param['name'])
            )

    def visitor_teardown(self):
        # implicit return for generator
        # PEP 380: return statement in generator is equivalent to raise StopIteration(value)
        # if not isinstance(self.opcodes[-1], (JavaOpcodes.ATHROW, JavaOpcodes.ARETURN)):
        # TODO: Uncomment the condition above once the issue described is resolved:
        # TODO: Currently need to throw StopIteration at the end of generator even when ATHROW/ARETURN is the last
        # TODO: instruction, to fix "'TRY' object has no attribute 'next_op'" error during transpilation
        self.add_opcodes(
            # StopIteration is a singleton by design, see org/python/exceptions/StopIteration
            JavaOpcodes.GETSTATIC(
                'org/python/exceptions/StopIteration', 'STOPITERATION', 'Lorg/python/exceptions/StopIteration;'
            ),
            JavaOpcodes.ATHROW(),
        )

    def transpile_method(self):
        return [
            JavaMethod(
                self.method_name + "$generator",
                '(Lorg/python/types/Generator;)Lorg/python/Object;',
                static=True,
                attributes=[self.transpile_code()] + self.method_attributes()
            )
        ]

    def transpile_wrapper(self):
        wrapper = Accumulator()

        wrapper.add_opcodes(
            # Construct a generator instance
            java.New('org/python/types/Generator'),

            # p1: The name of the generator
            JavaOpcodes.LDC_W(self.generator),

            # p2: The actual generator method
            java.Class.forName(self.class_descriptor.replace('/', '.')),

            JavaOpcodes.LDC_W(self.method_name + "$generator"),
            java.Array(1, 'java/lang/Class'),

            JavaOpcodes.DUP(),
            ICONST_val(0),
            java.Class.forName('org.python.types.Generator'),
            JavaOpcodes.AASTORE(),

            JavaOpcodes.INVOKEVIRTUAL(
                'java/lang/Class',
                'getMethod',
                args=['Ljava/lang/String;', '[Ljava/lang/Class;'],
                returns='Ljava/lang/reflect/Method;'
            ),

            # p3: The arguments passed to the generator method. These will be
            # restored on the first call to the generator.
            java.Map(),
        )

        for i, param in enumerate(self.parameters):
            wrapper.add_opcodes(
                JavaOpcodes.DUP(),
                JavaOpcodes.LDC_W(param['name']),

                JavaOpcodes.ALOAD(i + (0 if self.static else 1)),
                java.Map.put(),
            )

        if isinstance(self, GeneratorClosure):
            # stores a copy of closure variables
            wrapper.add_opcodes(
                JavaOpcodes.ALOAD_0(),  # first register contains initialized Closure object reference
                JavaOpcodes.CHECKCAST('org/python/types/Closure')
            )
            wrapper.add_opcodes(
                java.Init(
                    'org/python/types/Generator',
                    'Ljava/lang/String;',
                    'Ljava/lang/reflect/Method;',
                    'Ljava/util/Map;',
                    'Lorg/python/types/Closure;'
                ),
                JavaOpcodes.ARETURN(),
            )
        else:
            # Construct and return the generator object.
            wrapper.add_opcodes(
                java.Init(
                    'org/python/types/Generator',
                    'Ljava/lang/String;',
                    'Ljava/lang/reflect/Method;',
                    'Ljava/util/Map;'
                ),
                JavaOpcodes.ARETURN(),
            )

        return [
            JavaMethod(
                self.method_name,
                self.signature,
                static=self.static,
                attributes=[
                    JavaCode(
                        max_stack=len(self.parameters) + 9,
                        max_locals=len(self.parameters) + 8,
                        code=wrapper.opcodes
                    )
                ]
            )
        ]

    def store_name(self, name, declare=False):
        if declare or name in self.local_vars:
            self.add_opcodes(
                # Store in a local variable
                ASTORE_name(name),
            )
            self.add_opcodes(
                # Also store in the locals variable
                ALOAD_name('#locals'),
                JavaOpcodes.LDC_W(name),
                ALOAD_name(name),
                java.Map.put(),
            )
        else:
            # Unlike other Functions, GeneratorFunctions do not cache the current Module
            # locally, so it must be fetched on each use.
            self.add_opcodes(
                ASTORE_name('#value'),

                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),

                ALOAD_name('#value'),

                python.Object.set_attr(name),
                free_name('#value')
            )

    def load_name(self, name):
        if name == '<generator>':  # `<generator>` is not included in #locals
            self.add_opcodes(
                ALOAD_name(name)
            )
        elif name in self.local_vars:
            self.add_opcodes(
                ALOAD_name('#locals'),
                java.Map.get(name),
            )
        else:
            # Unlike other Functions, GeneratorFunctions do not cache the current Module
            # locally, so it must be fetched on each use.
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),
                python.Object.get_attribute(name),
            )

    def delete_name(self, name):
        try:
            self.add_opcodes(
                free_name(name),
                ALOAD_name('#locals'),
                java.Map.remove(name)
            )
        except NameError:
            # Unlike other Functions, GeneratorFunctions do not cache the current Module
            # locally, so it must be fetched on each use.
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),
                python.Object.del_attr(name),
            )


class GeneratorMethod(Method):
    def __init__(self, klass, name, code, generator, parameters, returns=None, static=False):
        super().__init__(
            klass,
            name=name,
            code=code,
            parameters=parameters,
            returns=returns,
            static=static,
        )
        self.generator = generator

    @property
    def klass(self):
        return self._parent

    def add_self(self):
        self.local_vars['<generator>'] = len(self.local_vars)
        self.has_self = True

    def visitor_setup(self):
        self.add_opcodes(
            # Restore the variables needed for the entry of the generator.
            ALOAD_name('<generator>'),
            JavaOpcodes.GETFIELD('org/python/types/Generator', 'stack', 'Ljava/util/Map;'),
            ASTORE_name('#locals'),
        )

        for param in self.parameters:
            self.add_opcodes(
                ALOAD_name('#locals'),
                java.Map.get(param['name']),
                ASTORE_name(param['name'])
            )

    def visitor_teardown(self):
        # implicit return for generator
        # PEP 380: return statement in generator is equivalent to raise StopIteration(value)
        # if not isinstance(self.opcodes[-1], (JavaOpcodes.ATHROW, JavaOpcodes.ARETURN)):
        # TODO: Uncomment the condition above once the issue described is resolved:
        # TODO: Currently need to throw StopIteration at the end of generator even when ATHROW/ARETURN is the last
        # TODO: instruction, to fix "'TRY' object has no attribute 'next_op'" error during transpilation
        self.add_opcodes(
            # StopIteration is a singleton by design, see org/python/exceptions/StopIteration
            JavaOpcodes.GETSTATIC(
                'org/python/exceptions/StopIteration', 'STOPITERATION', 'Lorg/python/exceptions/StopIteration;'
            ),
            JavaOpcodes.ATHROW(),
        )

    def transpile_method(self):
        return [
            JavaMethod(
                self.method_name + "$generator",
                '(Lorg/python/types/Generator;)Lorg/python/Object;',
                static=True,
                attributes=[self.transpile_code()] + self.method_attributes()
            )
        ]

    def transpile_wrapper(self):
        wrapper = Accumulator()

        wrapper.add_opcodes(
            # Construct a generator instance
            java.New('org/python/types/Generator'),

            # p1: The name of the generator
            JavaOpcodes.LDC_W(self.generator),

            # p2: The actual generator method
            java.Class.forName(self.klass.class_name),

            JavaOpcodes.LDC_W(self.method_name + "$generator"),
            java.Array(1, 'java/lang/Class'),

            JavaOpcodes.DUP(),
            ICONST_val(0),
            java.Class.forName('org.python.types.Generator'),
            JavaOpcodes.AASTORE(),

            JavaOpcodes.INVOKEVIRTUAL(
                'java/lang/Class',
                'getMethod',
                args=['Ljava/lang/String;', '[Ljava/lang/Class;'],
                returns='Ljava/lang/reflect/Method;'
            ),

            # p3: The arguments passed to the generator method. These will be
            # restored on the first call to the generator.
            java.Map(),
        )

        for i, param in enumerate(self.parameters):
            wrapper.add_opcodes(
                JavaOpcodes.DUP(),
                JavaOpcodes.LDC_W(param['name']),

                JavaOpcodes.ALOAD(i + (0 if self.static else 1)),
                java.Map.put(),
            )

        # Construct and return the generator object.
        wrapper.add_opcodes(
            java.Init(
                'org/python/types/Generator',
                'Ljava/lang/String;',
                'Ljava/lang/reflect/Method;',
                'Ljava/util/Map;',
            ),
            JavaOpcodes.ARETURN(),
        )

        return super().transpile_wrapper() + [
            JavaMethod(
                self.pyimpl_name,
                self.signature,
                static=self.static,
                attributes=[
                    JavaCode(
                        max_stack=len(self.parameters) + 9,
                        max_locals=len(self.parameters) + 8,
                        code=wrapper.opcodes
                    )
                ]
            )
        ]

    def store_name(self, name, declare=False):
        if declare or name in self.local_vars:
            self.add_opcodes(
                # Store in a local variable
                ASTORE_name(name),
            )
            self.add_opcodes(
                # Also store in the locals variable
                ALOAD_name('#locals'),
                JavaOpcodes.LDC_W(name),
                ALOAD_name(name),
                java.Map.put(),
            )
        else:
            # Unlike other Functions, GeneratorFunctions do not cache the current Module
            # locally, so it must be fetched on each use.
            self.add_opcodes(
                ASTORE_name('#value'),

                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),

                ALOAD_name('#value'),

                python.Object.set_attr(name),
                free_name('#value')
            )

    def load_name(self, name):
        if name == '<generator>':  # `<generator>` is not included in #locals
            self.add_opcodes(
                ALOAD_name(name)
            )
        else:
            super().load_name(name)

    def delete_name(self, name):
        try:
            self.add_opcodes(
                free_name(name)
            )
        except NameError:
            # Unlike other Functions, GeneratorFunctions do not cache the current Module
            # locally, so it must be fetched on each use.
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),
                python.Object.del_attr(name),
            )


class GeneratorClosure(GeneratorFunction):
    def __init__(self, module, code, generator, parameters, returns=None, static=False, outer_contexts=None):
        super().__init__(
            module,
            name='invoke',
            code=code,
            generator=generator,
            parameters=parameters,
            returns=returns,
            static=static,
        )
        self.nonlocal_vars = []  # holds nonlocal variable names for `store_name`
        self.outer_contexts = outer_contexts  # parent scopes of this closure, excluding global scope

    def __repr__(self):
        return '<GeneratorClosure %s (%s parameters)>' % (
            self.name, len(self.parameters)
        )

    @property
    def klass(self):
        return self._parent

    @property
    def module(self):
        return self.klass.module

    @property
    def class_descriptor(self):
        return self.klass.descriptor

    def store_name(self, name, declare=False):
        if name in self.nonlocal_vars:
            # updates closure
            self.add_opcodes(
                ALOAD_name('<generator>'),
                JavaOpcodes.CHECKCAST('org/python/types/Generator'),
                JavaOpcodes.GETFIELD('org/python/types/Generator', 'closure', 'Lorg/python/types/Closure;'),
                ICONST_val(_get_enclosing_context_level(self, name)),
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Closure',
                    'get_locals',
                    args=['I'],
                    returns='Ljava/util/Map;'
                ),
                JavaOpcodes.SWAP(),
                JavaOpcodes.LDC_W(name),
                JavaOpcodes.SWAP(),
                java.Map.put()
            )
        else:
            super().store_name(name, declare)

    def load_name(self, name):
        parent_level = _get_enclosing_context_level(self, name)

        if parent_level:
            self.add_opcodes(
                ALOAD_name('<generator>'),
                JavaOpcodes.CHECKCAST('org/python/types/Generator'),
                JavaOpcodes.GETFIELD('org/python/types/Generator', 'closure', 'Lorg/python/types/Closure;'),
                ICONST_val(parent_level),
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Closure',
                    'get_locals',
                    args=['I'],
                    returns='Ljava/util/Map;'
                ),
                java.Map.get(name),
            )
        else:
            super().load_name(name)



from ..utils import TranspileTestCase


class DeleteTests(TranspileTestCase):
    def test_delete_from_dict(self):
        self.assertCodeExecution("""
            x = {'a': 1, 'b': 2}

            print(" Delete 'a'...")
            del x['a']
            print('x =', x)

            try:
                print(" Delete 'a'...")
                del x['a']
                print("Key 'a' shouldn't exist")
            except KeyError:
                print("Key 'a' doesn't exist")
            print('x =', x)

            try:
                print(" Delete 'c'...")
                del x['c']
                print("Key 'c' never existed")
            except KeyError:
                print("Key 'c' doesn't exist")
            print('x =', x)

            print("Done.")
            """)

    def test_delete_from_list(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5, 6]

            print("x =", x)

            print("Delete index 2...")
            del x[2]
            print("x =", x)

            print("Delete index -1...")
            del x[-1]
            print("x =", x)

            try:
                print("Delete index 10...")
                del x[10]
                print("Index shouldn't exist")
            except IndexError:
                print("Index doesn't exist")

            print("x =", x)

            print("Done.")
            """)

    def test_delete_attribute(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self):
                    self.attr1 = 123

            obj = MyObject()

            obj.attr2 = 456
            print('obj.attr1 =', obj.attr1)
            print('obj.attr2 =', obj.attr2)

            print("Delete attributes...")
            del obj.attr1
            del obj.attr2

            print("Access deleted attributes...")
            try:
                print('obj.attr1 = ', obj.attr1)
                print("obj.attr1 shouldn't exist")
            except AttributeError:
                print("obj.attr1 doesn't exist")

            try:
                print('obj.attr2 = ', obj.attr2)
                print("obj.attr2 shouldn't exist")
            except AttributeError:
                print("obj.attr2 doesn't exist")

            print("Delete deleted attributes...")
            try:
                del obj.attr1
                print("obj.attr1 shouldn't existed")
            except AttributeError:
                print("obj.attr1 doesn't exist")

            try:
                del obj.attr2
                print("obj.attr2 shouldn't existed")
            except AttributeError:
                print("obj.attr2 doesn't exist")

            print("Delete non-existent attribute...")
            try:
                del obj.no_such_attr
                print("obj.no_such_attr never existed")
            except AttributeError:
                print("obj.no_such_attr doesn't exist")

            print("Done.")
            """)

    def test_delete_local(self):
        self.assertCodeExecution("""
            x = 123
            print('x = ', x)

            print('Deleting x...')
            del x
            try:
                print('x = ', x)
            except NameError:
                print('x was successfully deleted')

            def foo():
                y = 456
                print('y =', y)
                print('Deleting y...')
                del y
                try:
                    print('y =', y)
                except UnboundLocalError:
                    print('y was successfully deleted')

            foo()
            """)

    def test_delete_tuple(self):
        self.assertCodeExecution("""
            d = {"uno": 1, "due": 2, "tre": 3, "quattro": 4}
            print("keys before delete: ", sorted(d))
            del (d["uno"], d["quattro"])
            print("keys after delete: ", sorted(d))
            """)

    def test_delete_list(self):
        self.assertCodeExecution("""
            d = {"uno": 1, "due": 2, "tre": 3, "quattro": 4}
            print("keys before delete: ", sorted(d))
            del [d["uno"], d["quattro"]]
            print("keys after delete: ", sorted(d))
            """)
